/******************************************************************************/
/* BSeaMap -   BattleFleet  (impl.)            */
/******************************************************************************/
static const char bseamap_cpp_id[]="@(#) $Id: bseamap.cpp,v 1.20 2005/02/15 16:43:35 Oleksiy Balabay Exp $"; static const char* bseamap_cpp_id_id=bseamap_cpp_id;

#include <qfile.h>
#include <qdom.h>
#include <qregexp.h>
#include <math.h>
#include <qpainter.h>
#include <qtextcodec.h>

#include "bseamap.h"
#include "bapplication.h"
#include "bdomutils.h"
#include "bunitsconsts.h"

/** \file bseamap.cpp
  \brief  BSeaMap  BattleFleet (impl.)

     BSeaMap  BattleFleet
*/
const QString SM_MAP = "Map";
const QString SM_MAP_NAME = "Name";
const QString SM_GEOMETRY = "Geometry";
const QString SM_GEOMETRY_BOTTOM_LEFT = "BottomLeft";
const QString SM_GEOMETRY_TOP_RIGHT = "TopRight";
const QString SM_X = "X";
const QString SM_Y = "Y";
const QChar SM_DEGREES = 'D';
const QChar SM_MINUTES = 'M';
const QChar SM_SECONDS = 'S';
const QChar SM_LONGITUDE_EAST = 'E';
const QChar SM_LONGITUDE_WEST = 'W';
const QChar SM_LATITUDE_SOUTH = 'S';
const QChar SM_LATITUDE_NORTH = 'N';

const QString SM_ISLAND = "Island";
const QString SM_ISLAND_COLOR = "Color";
const QString SM_ISLAND_TYPE = "Type";
const QString SM_ISLAND_POINT = "Point";

const QString SM_ISLAND_TYPE_POLYGON = "Polygon";
const QString SM_ISLAND_TYPE_SPLINE = "Spline";

const QString SM_POINT = "Point";

const QString SM_ITEM_NAME = "Name";

const double BSeaMap::MIN_REGION = 32000;

/** \class BSeaMap
  \brief 

        ,  
    .     
   -   ,   ,   .

          .
  ,      ,  
  :
  <ul>
    <li>   X    Y , ..    
            
        </li>
    <li>       
    .        :
    <PRE>
       2*PI*EARTH_RADIUS_M*cos(latitude*2*PI/360)/(360*60*MILE);
    </PRE>
    </li>
    <li>     .  
            .
    </li>
  </ul>
  
       ,   . 
         
   .

       .    
   :
  <ul>
    <li>     \a  
    </li>
    <li> ( )    \a  
    </li>
    <li>   ,    </li>
    <li>   ,      
    </li>
  </ul>

          , 
      ,    
  ""  .

     :

  <BR><\a SM_MAP \a SM_MAP_NAME = "">
	<BR>&nbsp;&nbsp;<\a SM_GEOMETRY>
	<BR>&nbsp;&nbsp;&nbsp;&nbsp;<\a SM_GEOMETRY_BOTTOM_LEFT 
    \a SM_X="" \a SM_Y=""/>
	<BR>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<\a SM_GEOMETRY_TOP_RIGHT 
    \a SM_X="" \a SM_Y=""/>
	<BR>&nbsp;&nbsp;</ \a SM_GEOMETRY>
  <BR>&nbsp;&nbsp; 
  <BR>&nbsp;&nbsp;...
  <BR></ \a SM_MAP>

         degreesFromString().

        . :
  <ol>
    <li>BMapIsland - </li>
  </ol>

        background().

        
        - degreesFromString()  degreesToString().
          
    - degreesToSeconds()  degreesFromSeconds().

         originalCoordinates(),
     setOriginalCoordinates().
  
  ,        
     createTransformer()

            begin(), end().

*/

/** \brief 
    \a fileName   .   
   ,       
   BattleFleet.

  \exception BException     .     
    .
*/
BSeaMap::BSeaMap(const QString& fileName)
  :BXmlConfig(fileName), _coordTransformer(0) {
  try {
    parseDocument(*_doc);
    delete _doc;
    _doc = 0;
  }
  catch(BConfigException& e) {
    clearMap();
    throw addFileName(e);
  }
  catch (BException&) {
    clearMap();
    throw;
  }
  catch(...) {
    clearMap();
    throw BInternalException(__FILE__, __LINE__,
      qApp->translate("BXmlConfig",
      "Error loading file '%1'").arg(this->fileName()));
  }
}

/** \brief 
*/
BSeaMap::~BSeaMap(){
  clearMap();
}

void BSeaMap::clearMap() {
  for(iterator it = _mapItems.begin(); it != _mapItems.end(); it++) {
    delete *it;
  }
  _mapItems.clear();
//  if (_coordTransformer) {
//    delete _coordTransformer;
//    _coordTransformer = 0;
//  }
}

void BSeaMap::parseDocument(const QDomDocument& doc) {
  QDomElement docElem = doc.documentElement();
  if (docElem.tagName() != SM_MAP) {
    throw BConfigException(__FILE__, __LINE__,
      qApp->translate("BXmlConfig",
      "File '%1' is not '%2' file but '%3'").arg(fileName()).arg(
      SM_MAP).arg(docElem.tagName()));
  }
  bool geometryFound = false;
  QDomNode n = docElem.firstChild();
  while( !n.isNull() ) {
    QDomElement e = n.toElement(); // try to convert the node to an element.
    n = n.nextSibling();
    if( !e.isNull() ) { // the node was really an element.
      try {
        if (e.tagName() == SM_GEOMETRY) {
          parseGeometry(e);
          geometryFound = true;
        }
        else if (e.tagName() == SM_ISLAND) {
          BMapItem* item = new BMapIsland(e, this);
          _mapItems.push_back(item);
        }
        else {
          qWarning("Unsupported tag '%s' in file '%s'", 
            (const char*)e.tagName().local8Bit(),
            (const char*)fileName().local8Bit());
        }
      }
      catch (BConfigException& ex) {
        throw BConfigException(ex.sourceCodeFile(),
          ex.sourceCodeLine(), ex.message()+"\n"+qApp->translate("BSeaMap",
          "Error in element %1").arg(BDomUtils::elementToString(e)));
      }
      catch (BException&) {
        throw;
      }
      catch (...) {
        throw BInternalException(__FILE__, __LINE__,
          qApp->translate("BSeaMap",
          "Unknown error parsing element %1").arg(
          BDomUtils::elementToString(e)));
      }
    }
  }
  if (!geometryFound) {
    BDomUtils::elementDoesNotExist(SM_GEOMETRY);
  }
}

void BSeaMap::parseGeometry(const QDomElement& e) {
  if (e.tagName() != SM_GEOMETRY) {
    throw BConfigException(__FILE__, __LINE__,
      qApp->translate("BSeaMap", "Element '%1' expected, '%2' found").arg(
      SM_GEOMETRY).arg(e.tagName()));
  }
  QDomNode n = e.firstChild();
  double x0, y0, x1, y1;
  x0 = y0 = x1 = y1 = 0;
  bool bottomLeftFound = false;
  bool topRightFound = false;
  while( !n.isNull() ) {
    QDomElement e = n.toElement(); // try to convert the node to an element.
    n = n.nextSibling();
    if( !e.isNull() ) { // the node was really an element.
      if (e.tagName() == SM_GEOMETRY_BOTTOM_LEFT) {
        if ( (!e.hasAttribute(SM_X)) || (!e.hasAttribute(SM_Y))) {
          throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap", 
            "%1 element should have "
            "%2 and %3 attribute").arg(SM_GEOMETRY_BOTTOM_LEFT).arg(
            SM_X).arg(SM_Y));
        }
        x0 = degreesFromString(e.attribute(SM_X), longitude);
        y0 = degreesFromString(e.attribute(SM_Y), latitude);
        bottomLeftFound = true;
      }
      else if (e.tagName() == SM_GEOMETRY_TOP_RIGHT) {
        if ( (!e.hasAttribute(SM_X)) || (!e.hasAttribute(SM_Y))) {
          throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap", 
            "%1 element should have "
            "%2 and %3 attribute").arg(SM_GEOMETRY_TOP_RIGHT).arg(
            SM_X).arg(SM_Y));
        }
        x1 = degreesFromString(e.attribute(SM_X), longitude);
        y1 = degreesFromString(e.attribute(SM_Y), latitude);
        topRightFound = true;
      }
      else {
        qWarning("Unsupported tag '%s' in file '%s'", 
          (const char*)e.tagName().local8Bit(),
          (const char*)fileName().local8Bit());
      }
    }
  }
  if (!bottomLeftFound) {
    BDomUtils::elementDoesNotExist(SM_GEOMETRY+"::"+SM_GEOMETRY_BOTTOM_LEFT);
  }
  if (!topRightFound) {
    BDomUtils::elementDoesNotExist(SM_GEOMETRY+"::"+SM_GEOMETRY_TOP_RIGHT);
  }
  if (x0 > x1) {
    throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
      "Invalid geometry coordinates: %1 (%2) should be lower than %3 (%4)").arg(
      SM_GEOMETRY_BOTTOM_LEFT+"::"+SM_X).arg(degreesToString(x0, longitude, true)).arg(
      SM_GEOMETRY_TOP_RIGHT+"::"+SM_X).arg(degreesToString(x1, longitude, true)));
  }
  if (y0 > y1) {
    throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
      "Invalid geometry coordinates: %1 (%2) should be lower than %3 (%4)").arg(
      SM_GEOMETRY_BOTTOM_LEFT+"::"+SM_Y).arg(degreesToString(y0, latitude, true)).arg(
      SM_GEOMETRY_TOP_RIGHT+"::"+SM_Y).arg(degreesToString(y1, latitude, true)));
  }
  BCoordinates bc(x0, y0, x1 - x0, y1 - y0);
  setOriginalCoordinates(bc);
}

/** \brief       

   :   \a SM_DEGREES 
  [ \a SM_MINUTES [C \a SM_SECONDS]]

         , 
  .

    - .
*/
double BSeaMap::degreesFromString(const QString& degrees,
  const BCoordinatesType type) {
  QString s = degrees.stripWhiteSpace().upper();
  QChar h = s[0];
  //   
  int sign = 0;
  if (type == longitude) {
    if (h == SM_LONGITUDE_EAST) {
      sign = 1;
    }
    else if (h == SM_LONGITUDE_WEST) {
      sign = -1;
    }
    else {
      throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
        "Unknown hemisphere '%1'. '%2' or '%3' expected in '%4'").arg(
        h).arg(SM_LONGITUDE_EAST).arg(SM_LONGITUDE_WEST).arg(
        degrees));
    }
  }
  else if (type == latitude) {
    if (h == SM_LATITUDE_NORTH) {
      sign = 1;
    }
    else if (h == SM_LATITUDE_SOUTH) {
      sign = -1;
    }
    else {
      throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
        "Unknown hemisphere '%1'. '%2' or '%3' expected in '%4'"
        ).arg(h).arg(SM_LATITUDE_NORTH).arg(SM_LATITUDE_SOUTH).arg(
        degrees));
    }
  }
  //     
  s = s.mid(1).stripWhiteSpace();
  int endPos = s.find(QRegExp(QString("\\s*")+SM_DEGREES));
  if (endPos == -1) { //  
    throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
      "No '%1' found in '%2'").arg(SM_DEGREES).arg(degrees));
  }
  bool isOk = false;
  int deg = s.left(endPos).toUInt(&isOk);
  int minutes = 0;
  double seconds = 0;
  if (! isOk) {
    throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
      "Bad degrees '%1' found in '%2'").arg(s.left(endPos)).arg(
      degrees));
  }
  //     
  s = s.mid(endPos).stripWhiteSpace();
  Q_ASSERT(s[0] == SM_DEGREES);
  s = s.mid(1).stripWhiteSpace();
  endPos = s.find(QRegExp(QString("\\s*")+SM_MINUTES));
  if (endPos != -1) {
    minutes = s.left(endPos).toUInt(&isOk);
    if (! isOk) {
      throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
        "Bad minutes '%1' found in '%2'").arg(s.left(endPos)).arg(
        degrees));
    }
    //     
    s = s.mid(endPos).stripWhiteSpace();
    Q_ASSERT(s[0] == SM_MINUTES);
    s = s.mid(1).stripWhiteSpace();
  }
  endPos = s.find(QRegExp(QString("\\s*")+SM_SECONDS));
  if (endPos != -1) {
    seconds = s.left(endPos).toDouble(&isOk);
    if (! isOk) {
      throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
        "Bad seconds '%1' found in '%2'").arg(s.left(endPos)).arg(
        degrees));
    }
    s = s.mid(endPos).stripWhiteSpace();
    Q_ASSERT(s[0] == SM_SECONDS);
    s = s.mid(1).stripWhiteSpace();
  }
  // ,     
  if (s.length() > 0) {
    throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
      "Unknown data after seconds sign '%1' found in '%2'").arg(
      s.left(endPos)).arg(degrees));
  }
  // ToDo:  .
  if ((type == longitude) && ( (deg < 0) || (deg >180) ) ) {
    throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
      "Degrees should be from 0 to 180. "
      "'%1' found in '%2'").arg(deg).arg(
      degrees));
  }
  if ((type == longitude) && (deg == 180) && ((minutes >0)||(seconds >0)) ) {
    throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
      "Degrees should be from 0 to 180. "
      "'%1' found in '%2'").arg(deg).arg(
      degrees));
  }
  if ((type == latitude) && ( (deg < 0) || (deg >90) ) ) {
    throw BConfigException(__FILE__, __LINE__, qApp->translate("BSeaMap",
      "Degrees should be from 0 to 90. "
      "'%1' found in '%2'").arg(deg).arg(
      degrees));
  }
  double dgrs = sign * (deg + minutes/60.0 + seconds/(60.0*60.0));
  return dgrs;
}

/** \brief    

   \a hrf == \a true (hrf - human readable form),
     '  ",  
,   degreesFromString()

*/
QString BSeaMap::degreesToString(const double degrees, 
  const BCoordinatesType type, const bool hrf) {
  int deg; int m; int is; double s;
  BUnitsConverter::angle(degrees, &deg, &m, &is, &s, 0);
  //  
  QString hemisphere;
  QChar nonLocalizedHemisphere;
  if (latitude == type) {
    if (degrees>=0) {
      hemisphere = qApp->translate("BSeaMap", "N", "Norsern");
      nonLocalizedHemisphere = SM_LATITUDE_NORTH;
    }
    else {
      hemisphere = qApp->translate("BSeaMap", "S", "Southern");
      nonLocalizedHemisphere = SM_LATITUDE_SOUTH;
    }
  }
  else { //longitude
    if (degrees>=0) {
      hemisphere = qApp->translate("BSeaMap", "E", "Eastern");
      nonLocalizedHemisphere = SM_LONGITUDE_EAST;
    }
    else {
      hemisphere = qApp->translate("BSeaMap", "W", "Western");
      nonLocalizedHemisphere = SM_LONGITUDE_WEST;
    }
  }
  if (hrf) {
    return QString("%1%2%3%4%5%6%7").arg(deg).arg(QChar(176)).arg(
      m).arg("'").arg(
      is).arg("\"").arg(
      hemisphere);
  }
  else {
    return QString("%1 %2%3 %4%5 %6%7").arg(nonLocalizedHemisphere).arg(
      deg).arg(SM_DEGREES).arg(
      m).arg(SM_MINUTES).arg(
      s).arg(SM_SECONDS);
  }
}

int BSeaMap::degreesToSeconds(const double degrees) {
  int sign = (degrees >= 0)?1:-1;
  double d = fabs(degrees);
  int secs = static_cast<int>(floor(d*60*60));
  return secs*sign;
}

double BSeaMap::degreesFromSeconds(const int seconds) {
  return seconds/(60.0*60);
}

int BSeaMap::secondsPart(const double degrees) {
  int sign = (degrees >= 0)?1:-1;
  int deg = degreesPart(degrees);
  int m_s = static_cast<int>((fabs(degrees) - abs(deg))*60*60);
  return m_s*sign%60;
}

int BSeaMap::minutesPart(const double degrees) {
  int sign = (degrees >= 0)?1:-1;
  int deg = degreesPart(degrees);
  int m_s = static_cast<int>((fabs(degrees) - abs(deg))*60*60);
  return m_s*sign/60;
}

int BSeaMap::degreesPart(const double degrees) {
  return degreesToSeconds(degrees)/(60*60);
}

double BSeaMap::degrees(const int degrees, const int minutes, 
  const int seconds) {
  return degrees + minutes/60.0 + seconds/(60.0*60);
}

/** \brief      canvas

  setRegion()     
*/
QPointArray BCoordTransformer::shape(const BMapPoints& mapPoints) const {
  QPointArray pa(mapPoints.size());
  typedef BMapPoints::const_iterator CIT;
  int i = 0;
  for (CIT it = mapPoints.begin(); it != mapPoints.end(); it++, i++) {
    double newX, newY;
    _currentMatrix.map(it->first, it->second, &newX, &newY);
    pa.setPoint(i, newX, newY);
  }
//  for (i = 0; i < pa.size(); i++) {
//    qDebug("%d x=%d y=%d", i, pa[i].x(), pa[i].y());
//  }
  return pa;
}

BSeaMap::iterator BSeaMap::begin() {
  return _mapItems.begin();
}

BSeaMap::iterator BSeaMap::end() {
  return _mapItems.end();
}

BSeaMap::const_iterator BSeaMap::begin() const {
  return _mapItems.begin();
}

BSeaMap::const_iterator BSeaMap::end() const {
  return _mapItems.end();
}

const BCoordinates BCoordTransformer::originalCoordinates() const {
  return _co;
}

const BCoordinates BCoordTransformer::currentCoordinates() const {
  return _mapCo;
}

const BCoordinates BCoordTransformer::currentCanvasCoordinates() const {
  return _canvasCo;
}

/** \brief      canvas
*/
QPoint BCoordTransformer::pointFromMap(const BMapPoint& point) const {
  double newX, newY;
  _currentMatrix.map(point.first, point.second, &newX, &newY);
  return QPoint(newX, newY);
}

/** \brief   canvas   
*/
BMapPoint BCoordTransformer::pointToMap(const QPoint& point) const {
  double newX, newY;
  _currentMatrix.invert().map(point.x(), point.y(), &newX, &newY);
  return BMapPoint(newX, newY);
}

int BCoordTransformer::sizeFromMap(const double miles) const {
  double oY = _co.y0();
  double newY, newY1, tmp;
  _currentMatrix.map(0, oY, &tmp, &newY);
  _currentMatrix.map(0, oY+ BUnitsConverter::degreesFromMiles(miles), 
    &tmp, &newY1);
  int sign = (miles > 0)?1:-1;
  return abs(newY1 - newY)*sign;
}

double BCoordTransformer::sizeToMap(const int pixels) const {
  double oY = _co.y0();
  double newY, newY1, tmp;
  _currentMatrix.invert().map(0, oY, &tmp, &newY);
  _currentMatrix.invert().map(0, oY+pixels, &tmp, &newY1);
  int sign = (pixels > 0)?1:-1;
  return fabs(BUnitsConverter::degreesToMiles(newY1 - newY))*sign;
}

double BCoordTransformer::minRegion(const BCoordinates& /* canvasCo */) const {
//  const BCoordinates* origCo = originalCoordinates();
//  double minRegionX = origCo->w()*canvasCo.w()/MIN_REGION;
//  double minRegionY = origCo->h()*canvasCo.h()/MIN_REGION;
//  return QMAX(minRegionX, minRegionY);
  return 1/120.0;
}
/**
  ""         
*/
BCoordinates BCoordTransformer::adjustMapCoordinates(const BCoordinates& mapCo,
  const BCoordinates& canvasCo) const {
  const BCoordinates origCo = originalCoordinates();
  double lk = BUnitsConverter::longitudeCoef(_co.y0());
  //        ,    
  double x0 = mapCo.x0();
  if (x0 < origCo.x0())
    x0 = origCo.x0();
  double y0 = mapCo.y0();
  if (y0 < origCo.y0())
    y0 = origCo.y0();
  //        ,    
  double x1 = mapCo.x0() + mapCo.w();
  if (x1 > origCo.x0() + origCo.w()) {
    x1 = origCo.x0() + origCo.w();
  }
  double y1 = mapCo.y0() + mapCo.h();
  if (y1 > origCo.y0() + origCo.h()) {
    y1 = origCo.y0() + origCo.h();
  }
  double minR = minRegion(canvasCo);
  double minRX = minR*lk;
  //     > ,     ,  
  //  
  if (x0 > x1) {
    x0 = x1 - minRX;
  }
  if (y0 > y1) {
    y0 = y1 - minR;
  }
  //   
  if (x1 - x0 < minRX) {
    //   
    x1 = x0 + minRX;
  }
  if (y1 - y0 < minR) {
    //   
    y0 = y1 - minR;
  }
  // TODO:  

  double w = (x1 - x0)*lk;
  double h = (y1 - y0);
//  float k1 = canvasCo.w()/(w);
//  float k2 = canvasCo.h()/h;
//  double w = mapCo.w()*lk;
//  double h = mapCo.h();
//  float k1 = canvasCo.w()/(mapCo.w()*lk);
//  float k2 = canvasCo.h()/mapCo.h();
  //  
//  if (k1 < k2) {
//   h = canvasCo.h()*(w)/canvasCo.w();
//  }
//  else {
//    w = canvasCo.w()*h/(canvasCo.h());
//  }
  if (h*canvasCo.w()/canvasCo.h() > w) {
    w = h*canvasCo.w()/canvasCo.h();
  }
  else {
    h = w*canvasCo.h()/canvasCo.w();
  }
  return BCoordinates(x0, y0, w/lk, h);
}

BMapItem::BMapItem(const QDomElement& /* e */, const BSeaMap* seaMap) 
 :_seaMap(seaMap) {
}

BMapItem::~BMapItem() {
}

QString BMapItem::name() const {
  return _name;
}

void BMapItem::setName(const QString& name) {
  _name = name;
}

const BSeaMap* BMapItem::seaMap() const {
  return _seaMap;
}

/** \brief    Xml  

   Xml -  BSeaMap::degreesFromString().
  Xml - <\a SM_POINT \a SM_X="X" \a SM_Y="Y" />.
   .

*/
BMapPoints BMapItem::mapPointsFromXml(const QDomElement& e)
 {
  BMapPoints mapPoints;
  QDomNodeList nl = e.elementsByTagName(SM_POINT);
  for (int i = 0; i < static_cast<int>(nl.count()); i++) {
    QDomElement e = nl.item(i).toElement();
    if ((!e.hasAttribute(SM_X)) || (!e.hasAttribute(SM_Y))) {
      throw BConfigException(__FILE__, __LINE__,qApp->translate("BMapItem",
        "%1 should have attributes %2 and %3. Error in element %4").arg(
        SM_POINT).arg(SM_X).arg(SM_Y).arg(BDomUtils::elementToString(e)));
    }
    double x=BSeaMap::degreesFromString(e.attribute(SM_X), BSeaMap::longitude);
    double y=BSeaMap::degreesFromString(e.attribute(SM_Y), BSeaMap::latitude);
    mapPoints.push_back(BMapPoint(x,y));
  }
  return mapPoints;
}


BMapIsland::BMapIsland(const QDomElement& e, const BSeaMap* seaMap) 
 :BMapItem(e, seaMap) {
  _type = islandTypeFromString(e.attribute(SM_ISLAND_TYPE,
    SM_ISLAND_TYPE_POLYGON));
  _color = BDomUtils::colorFromString(e.attribute(SM_ISLAND_COLOR, "black"));
  _points = mapPointsFromXml(e);
  setName(e.attribute(SM_ITEM_NAME, ""));
  if (_points.size() == 0) {
    throw BConfigException(__FILE__, __LINE__,
      qApp->translate("BMapIsland",
      "No points found for island %1").arg(name()));
  }
}

BMapIsland::~BMapIsland() {
}

/* QCanvasItem* BMapIsland::createCanvasItem(QCanvas* canvas) {
  if (PolygonType == _type) {
    QCanvasPolygon* ci = new QCanvasPolygon(canvas);
    QPointArray pa = seaMap()->shape(_points);
    ci->setPoints(pa);
    ci->setBrush(QBrush(_color));
    return ci;
  }
  else if (SplineType == _type) {
    QCanvasSpline* ci = new QCanvasSpline(canvas);
    ci->setControlPoints(seaMap()->shape(_points), true);
    ci->setBrush(QBrush(_color));
    return ci;
  }
  return 0;
}
*/

void BMapIsland::drawBackground(const BCoordTransformer& coordTransformer,
  QPainter& p) {
  if (PolygonType == _type) {
    QPointArray pa = coordTransformer.shape(_points);
    p.setBrush(QBrush(_color));
    p.drawPolygon(pa, true);
  }
}


BMapIsland::BIslandType BMapIsland::islandTypeFromString(
  const QString& islandType) {
  if (SM_ISLAND_TYPE_POLYGON == islandType) {
    return PolygonType;
  }
  else if (SM_ISLAND_TYPE_SPLINE == islandType) {
    return SplineType;
  }
  else {
    throw BConfigException(__FILE__, __LINE__,
      qApp->translate("BMapIsland", 
      "Invalid island attribute %1: '%2'. "
      "Allowed values: '%3', '%4'").arg(SM_ISLAND_TYPE).arg(
      islandType).arg(SM_ISLAND_TYPE_POLYGON).arg(
      SM_ISLAND_TYPE_SPLINE));
  }
}

QPixmap BSeaMap::background(const BCoordTransformer& coordTransformer) const {
  QPixmap pm(coordTransformer.canvasSize());
  pm.fill();
  {
    BCoordinates canvasCo = coordTransformer.currentCanvasCoordinates();
    QPainter p(&pm);
    p.setClipping(true);
    p.setClipRect(canvasCo.x0(), canvasCo.y0(), 
      canvasCo.w(), canvasCo.h(), QPainter::CoordPainter);
    for (const_iterator it = begin(); it != end(); it++) {
      (*it)->drawBackground(coordTransformer, p);
    }
  }
  return pm;
}

/** \class BCoordTransformer

  \brief        
*/

/** 
  \param origCo -    
  \param mapCo -       
     
  \param canvasCo -     canvas',  
    .  
  \param canvasSize -    

*/
BCoordTransformer::BCoordTransformer(const BCoordinates& origCo,
  const BCoordinates& mapCo,
  const BCoordinates& canvasCo,
  const QSize& canvasSize):_co(origCo), _mapCo(mapCo), 
  _canvasCo(canvasCo), _canvasSize(canvasSize) {
  QWMatrix m1;
  m1.translate(_mapCo.x0(), _mapCo.y0());
  //   
  m1.scale(1/BUnitsConverter::longitudeCoef(_co.y0()), 1);
  m1.scale(/*_mapCo->w()/_canvasCo->w()*/
    _mapCo.h()/_canvasCo.h(), _mapCo.h()/_canvasCo.h());
  QWMatrix m2(1,0,0,-1,0,canvasSize.height());
  QWMatrix m3 = m2*m1*_originalMatrix;
  m3.translate(-_canvasCo.x0(), _canvasCo.y0());
  _currentMatrix = (m3).invert();
}

QSize BCoordTransformer::canvasSize() const {
  return _canvasSize;
}

BCoordTransformer* BSeaMap::createTransformer(const BCoordinates& mapCo, 
  const BCoordinates& canvasCo, const QSize& canvasSize) {
  BCoordinates mapCoAdjusted = 
    _coordTransformer->adjustMapCoordinates(mapCo, canvasCo);

  return new BCoordTransformer(_coordTransformer->originalCoordinates(),
    mapCoAdjusted, canvasCo, canvasSize);
}

const BCoordinates BSeaMap::originalCoordinates() const {
  return _coordTransformer->originalCoordinates();
}

QDomDocument BSeaMap::toXml() const {
  QDomDocument doc;
  BCoordinates co = originalCoordinates();
  QDomProcessingInstruction pi = doc.createProcessingInstruction(
    "xml", QString("version='1.0' encoding='UTF-8'"));
  doc.insertBefore(pi, doc.documentElement());
  doc.appendChild(doc.createElement(SM_MAP));
  QDomElement geometryEl = doc.createElement(SM_GEOMETRY);
  doc.documentElement().appendChild(geometryEl);

  {
    QDomElement bottomLeftEl = doc.createElement(SM_GEOMETRY_BOTTOM_LEFT);
    bottomLeftEl.setAttribute(SM_X, 
      degreesToString(co.x0(), longitude));
    bottomLeftEl.setAttribute(SM_Y, 
      degreesToString(co.y0(), latitude));
    geometryEl.appendChild(bottomLeftEl);
  }

  {
    QDomElement topRightEl = doc.createElement(SM_GEOMETRY_TOP_RIGHT);
    topRightEl.setAttribute(SM_X, 
      degreesToString(co.x0()+co.w(), longitude));
    topRightEl.setAttribute(SM_Y, 
      degreesToString(co.y0() + co.h(), latitude));
    geometryEl.appendChild(topRightEl);
  }
  return doc;
}

void BSeaMap::setOriginalCoordinates(const BCoordinates& original) {
  BCoordinates cc(0, 0, 100, 100);
  _coordTransformer = std::auto_ptr<BCoordTransformer>(new BCoordTransformer(
    original, original, cc, QSize(100,100)));
}

double BSeaMap::mileage2(const double x1, const double y1, 
  const double x2, const double y2) {
  double y0 = originalCoordinates().y0();
  double x11 = BUnitsConverter::degreesToMiles(x1, y0);
  double y11 = BUnitsConverter::degreesToMiles(y1);
  double x21 = BUnitsConverter::degreesToMiles(x2, y0);
  double y21 = BUnitsConverter::degreesToMiles(y2);
  return (x21-x11)*(x21-x11) + (y21-y11)*(y21-y11);
}

double BSeaMap::mileage(const double x1, const double y1, 
  const double x2, const double y2) {
  return sqrt( mileage2(x1, y1, x2, y2) );
}

BCoordinates::BCoordinates(double x0, double y0, double w, double h)
  :_x0(x0), _y0(y0), _w(w), _h(h) {
  if (w < 0) {
    _x0 = _x0 + _w;
    _w = fabs(_w);
  }
  if (h < 0) {
    _y0 = _y0 + _h;
    _h = fabs(_h);
  }
  if (_x0 < -180) {
    _x0 = -180;
  }
  if (_y0 < -90) {
    _y0 = -90;
  }
}
