/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Designer.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding valid Qt Enterprise Edition or Qt Professional Edition
** licenses may use this file in accordance with the Qt Commercial License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about Qt Commercial License Agreements.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#include <qvariant.h> // HP-UX compiler needs this here

#include "propertyeditor.h"
#include "globaldefs.h"
#include "defs.h"
#include "asciivalidator.h"
#include <qpopupmenu.h>
#include <qpainter.h>
#include <qpalette.h>
#include <qapplication.h>
#include <qheader.h>
#include <qlineedit.h>
#include <qstrlist.h>
#include <qmetaobject.h>
#include <qcombobox.h>
#include <qpushbutton.h>
#include <qhbox.h>
#include <qfontdialog.h>
#include <qspinbox.h>
#include <qevent.h>
#include <qobjectlist.h>
#include <qlistbox.h>
#include <qfontdatabase.h>
#include <qcolor.h>
#include <qcolordialog.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qsizepolicy.h>
#include <qbitmap.h>
#include <qtooltip.h>
#include <qwhatsthis.h>
#include <qaccel.h>
#include <qworkspace.h>
#include <qtimer.h>
#include <qdragobject.h>
#include <qdom.h>
#include <qprocess.h>
#include <qstyle.h>
#include <qdatetimeedit.h>
#include <qassistantclient.h>
#include <qdrawutil.h>
#include <qmultilineedit.h> // FIXME: obsolete
#include <qsplitter.h>
#include <qdatatable.h>
#include <qtextview.h>

#include <limits.h>

#include "bpropertyitems.h"

static QFontDatabase *fontDataBase = 0;
QString assistantPath();

#ifdef Q_WS_MAC
#include <qwindowsstyle.h>
static void setupStyle(QWidget *w)
{
  static QStyle *windowsStyle = 0;
  if(!windowsStyle)
    windowsStyle = new QWindowsStyle;
  w->setStyle(windowsStyle);
}
#else
#define setupStyle(x)
#endif

static void cleanupFontDatabase()
{
  delete fontDataBase;
  fontDataBase = 0;
}

static QStringList getFontList()
{
  if ( !fontDataBase ) {
    fontDataBase = new QFontDatabase;
    qAddPostRoutine( cleanupFontDatabase );
  }
  return fontDataBase->families();
}


class PropertyWhatsThis : public QWhatsThis
{
public:
  PropertyWhatsThis( PropertyList *l );
  QString text( const QPoint &pos );
  bool clicked( const QString& href );
  
private:
  PropertyList *propertyList;
  
};

PropertyWhatsThis::PropertyWhatsThis( PropertyList *l )
: QWhatsThis( l->viewport() ), propertyList( l )
{
}

QString PropertyWhatsThis::text( const QPoint &pos )
{
  return propertyList->whatsThisAt( pos );
}

bool PropertyWhatsThis::clicked( const QString& href )
{
  if ( !href.isEmpty() ) {
    //	QAssistantClient *ac = MainWindow::self->assistantClient();
    //	ac->showPage( QString( qInstallPathDocs() ) + "/html/" + href );
  }
  return FALSE; // do not hide window
}


/*!
\class PropertyItem propertyeditor.h
\brief Base class for all property items

  This is the base class for each property item for the
  PropertyList. A simple property item has just a name and a value to
  provide an editor for a datatype. But more complex datatypes might
  provide an expandable item for editing single parts of the
  datatype. See hasSubItems(), initChildren() for that.
*/

/*!  If this item should be a child of another property item, specify
\a prop as the parent item.
*/

PropertyItem::PropertyItem( PropertyList *l, PropertyItem *after, PropertyItem *prop, const QString &propName )
: QListViewItem( l, after ), listview( l ), property( prop ), propertyName( propName )
{
  setEditable(false);
  setSelectable( FALSE );
  open = FALSE;
  setText( 0, propertyName );
  changed = FALSE;
  setText( 1, "" );
  resetButton = 0;
}

PropertyItem::~PropertyItem()
{
  if ( resetButton )
    delete resetButton->parentWidget();
  resetButton = 0;
}

void PropertyItem::toggle()
{
}

void PropertyItem::updateBackColor()
{
  if ( itemAbove() && this != listview->firstChild() ) {
    if ( ( ( PropertyItem*)itemAbove() )->backColor == *backColor1 )
      backColor = *backColor2;
    else
      backColor = *backColor1;
  } else {
    backColor = *backColor1;
  }
  if ( listview->firstChild() == this )
    backColor = *backColor1;
}

QColor PropertyItem::backgroundColor()
{
  updateBackColor();
  if ( (QListViewItem*)this == listview->currentItem() )
    return *selectedBack;
  return backColor;
}

/*!  If a subclass is a expandable item, this is called when the child
items should be created.
*/

void PropertyItem::createChildren()
{
}

/*!  If a subclass is a expandable item, this is called when the child
items should be initialized.
*/

void PropertyItem::initChildren()
{
}

void PropertyItem::paintCell( QPainter *p, const QColorGroup &cg, int column, int width, int align )
{
  QColorGroup g( cg );
  g.setColor( QColorGroup::Base, backgroundColor() );
  g.setColor( QColorGroup::Foreground, Qt::black );
  g.setColor( QColorGroup::Text, Qt::black );
  int indent = 0;
  if ( column == 0 ) {
    indent = 20 + ( property ? 20 : 0 );
    p->fillRect( 0, 0, width, height(), backgroundColor() );
    p->save();
    p->translate( indent, 0 );
  }
  
  if ( isChanged() && column == 0 ) {
    p->save();
    QFont f = p->font();
    f.setBold( TRUE );
    p->setFont( f );
  }
  
  if ( !hasCustomContents() || column != 1 ) {
    QListViewItem::paintCell( p, g, column, width - indent, align  );
  } else {
    p->fillRect( 0, 0, width, height(), backgroundColor() );
    drawCustomContents( p, QRect( 0, 0, width, height() ) );
  }
  
  if ( isChanged() && column == 0 )
    p->restore();
  if ( column == 0 )
    p->restore();
  if ( hasSubItems() && column == 0 ) {
    p->save();
    p->setPen( cg.foreground() );
    p->setBrush( cg.base() );
    p->drawRect( 5, height() / 2 - 4, 9, 9 );
    p->drawLine( 7, height() / 2, 11, height() / 2 );
    if ( !isOpen() )
      p->drawLine( 9, height() / 2 - 2, 9, height() / 2 + 2 );
    p->restore();
  }
  p->save();
  p->setPen( QPen( cg.dark(), 1 ) );
  p->drawLine( 0, height() - 1, width, height() - 1 );
  p->drawLine( width - 1, 0, width - 1, height() );
  p->restore();
  
  if ( listview->currentItem() == this && column == 0 &&
    !listview->hasFocus() && !listview->viewport()->hasFocus() )
    paintFocus( p, cg, QRect( 0, 0, width, height() ) );
}

void PropertyItem::paintBranches( QPainter * p, const QColorGroup & cg,
                                 int w, int y, int h )
{
  QColorGroup g( cg );
  g.setColor( QColorGroup::Base, backgroundColor() );
  QListViewItem::paintBranches( p, g, w, y, h );
}

void PropertyItem::paintFocus( QPainter *p, const QColorGroup &cg, const QRect &r )
{
  p->save();
  QApplication::style().drawPrimitive(QStyle::PE_Panel, p, r, cg,
    QStyle::Style_Sunken, QStyleOption(1,1) );
  p->restore();
}

/*!  Subclasses which are expandable items have to return TRUE
here. Default is FALSE.
*/

bool PropertyItem::hasSubItems() const
{
  return FALSE;
}

/*!  Returns the parent property item here if this is a child or 0
otherwise.
*/

PropertyItem *PropertyItem::propertyParent() const
{
  return property;
}

bool PropertyItem::isOpen() const
{
  return open;
}

void PropertyItem::setOpen( bool b )
{
  if ( b == open )
    return;
  open = b;
  
  if ( !open ) {
    children.setAutoDelete( TRUE );
    children.clear();
    children.setAutoDelete( FALSE );
    qApp->processEvents();
    listview->updateEditorSize();
    return;
  }
  
  createChildren();
  initChildren();
  qApp->processEvents();
  listview->updateEditorSize();
}

/*!  Subclasses have to show the editor of the item here
*/

void PropertyItem::showEditor()
{
  createResetButton();
  resetButton->parentWidget()->show();
}

/*!  Subclasses have to hide the editor of the item here
*/

void PropertyItem::hideEditor()
{
  createResetButton();
  resetButton->parentWidget()->hide();
}

/*!  This is called to init the value of the item. Reimplement in
subclasses to init the editor
*/

void PropertyItem::setValue( const QVariant &v )
{
  val = v;
}

QVariant PropertyItem::value() const
{
  return val;
}

bool PropertyItem::isChanged() const
{
  return changed;
}

void PropertyItem::setChanged( bool b, bool updateDb )
{
  if ( propertyParent() )
    return;
  if ( changed == b )
    return;
  changed = b;
  repaint();
  if ( updateDb ) {
    //	MetaDataBase::setPropertyChanged( listview->propertyEditor()->widget(), name(), changed );
  }
  updateResetButtonState();
}

QString PropertyItem::name() const
{
  return propertyName;
}

void PropertyItem::createResetButton()
{
  if ( resetButton ) {
    resetButton->parentWidget()->lower();
    return;
  }
  QHBox *hbox = new QHBox( listview->viewport() );
  hbox->setFrameStyle( QFrame::StyledPanel | QFrame::Sunken );
  hbox->setLineWidth( 1 );
  resetButton = new QPushButton( hbox );
  setupStyle( resetButton );
  resetButton->setPixmap( QPixmap::fromMimeSource( "designer_resetproperty.png" ) );
  resetButton->setFixedWidth( resetButton->sizeHint().width() );
  hbox->layout()->setAlignment( Qt::AlignRight );
  listview->addChild( hbox );
  hbox->hide();
  QObject::connect( resetButton, SIGNAL( clicked() ),
    listview, SLOT( resetProperty() ) );
  QToolTip::add( resetButton, QObject::tr( "Reset the property to its default value" ) );
  QWhatsThis::add( resetButton, QObject::tr( "Click this button to reset the property to its default value" ) );
  updateResetButtonState();
}

void PropertyItem::updateResetButtonState()
{
  if ( !resetButton )
    return;
  if ( propertyParent() /*|| !WidgetFactory::canResetProperty( listview->propertyEditor()->widget(), name() )*/ )
    resetButton->setEnabled( FALSE );
  else
    resetButton->setEnabled( isChanged() );
}

/*!  Call this to place/resize the item editor correctly (normally
call it from showEditor())
*/

void PropertyItem::placeEditor( QWidget *w )
{
  createResetButton();
  QRect r = listview->itemRect( this );
  if ( !r.size().isValid() ) {
    listview->ensureItemVisible( this );
#if defined(Q_WS_WIN)
    listview->repaintContents( FALSE );
#endif
    r = listview->itemRect( this );
  }
  r.setX( listview->header()->sectionPos( 1 ) );
  r.setWidth( listview->header()->sectionSize( 1 ) - 1 );
  r.setWidth( r.width() - resetButton->width() - 2 );
  r = QRect( listview->viewportToContents( r.topLeft() ), r.size() );
  w->resize( r.size() );
  listview->moveChild( w, r.x(), r.y() );
  resetButton->parentWidget()->resize( resetButton->sizeHint().width() + 10, r.height() );
  listview->moveChild( resetButton->parentWidget(), r.x() + r.width() - 8, r.y() );
  resetButton->setFixedHeight( QMAX( 0, r.height() - 3 ) );
}

/*!  This should be called by subclasses if the user changed the value
of the property and this value should be applied to the widget property
*/

void PropertyItem::notifyValueChange()
{
  if ( !propertyParent() ) {
    listview->valueChanged( this );
    setChanged( TRUE );
    if ( hasSubItems() )
      initChildren();
  } else {
    propertyParent()->childValueChanged( this );
    setChanged( TRUE );
  }
}

/*!  If a subclass is a expandable item reimplement this as this is
always called if a child item changed its value. So update the
display of the item here then.
*/

void PropertyItem::childValueChanged( PropertyItem * )
{
}

/*!  When adding a child item, call this (normally from addChildren()
*/

void PropertyItem::addChild( PropertyItem *i )
{
  children.append( i );
}

int PropertyItem::childCount() const
{
  return children.count();
}

PropertyItem *PropertyItem::child( int i ) const
{
  // ARRRRRRRRG
  return ( (PropertyItem*)this )->children.at( i );
}

/*!  If the contents of the item is not displayable with a text, but
you want to draw it yourself (using drawCustomContents()), return
TRUE here.
*/

bool PropertyItem::hasCustomContents() const
{
  return FALSE;
}

/*!
\sa hasCustomContents()
*/

void PropertyItem::drawCustomContents( QPainter *, const QRect & )
{
}

QString PropertyItem::currentItem() const
{
  return QString::null;
}

int PropertyItem::currentIntItem() const
{
  return -1;
}

void PropertyItem::setCurrentItem( const QString & )
{
}

void PropertyItem::setCurrentItem( int )
{
}

int PropertyItem::currentIntItemFromObject() const
{
  return -1;
}

QString PropertyItem::currentItemFromObject() const
{
  return QString::null;
}

void PropertyItem::setFocus( QWidget *w )
{
if ( !qApp->focusWidget() /*||
                          listview->propertyEditor()->formWindow() &&
                          ( !MainWindow::self->isAFormWindowChild( qApp->focusWidget() ) &&
                          !qApp->focusWidget()->inherits( "Editor" ) ) */)
                          w->setFocus();
}

void PropertyItem::setText( int col, const QString &t )
{
  QString txt( t );
  if ( col == 1 )
    txt = txt.replace( "\n", " " );
  QListViewItem::setText( col, txt );
}

const QString& PropertyItem::whatsThisText() const {
  return _whatsThisText;
}

void PropertyItem::setWhatsThisText(const QString& whatsThisTxt) {
  _whatsThisText = whatsThisTxt;
  for(int i = 0; i < childCount(); i++) {
    child(i)->setWhatsThisText(whatsThisTxt);
  }
}
// --------------------------------------------------------------

PropertyTextItem::PropertyTextItem( PropertyList *l, PropertyItem *after, PropertyItem *prop,
                                   const QString &propName, bool comment, bool multiLine, bool ascii, bool a,
                                   bool ro): PropertyItem( l, after, prop, propName ), withComment( comment ),
                                   hasMultiLines( multiLine ), asciiOnly( ascii ), accel( a ), _readOnly(ro)
{
  lin = 0;
  box = 0;
}

QLineEdit *PropertyTextItem::lined()
{
  if ( lin )
    return lin;
  if ( hasMultiLines ) {
    box = new QHBox( listview->viewport() );
    box->setFrameStyle( QFrame::StyledPanel | QFrame::Sunken );
    box->setLineWidth( 2 );
    box->hide();
  }
  
  lin = 0;
  if ( hasMultiLines )
    lin = new QLineEdit( box );
  else
    lin = new QLineEdit( listview->viewport() );
  lin->setReadOnly(_readOnly);
  if ( asciiOnly ) {
    if ( PropertyItem::name() == "name" ) {
      lin->setValidator( new AsciiValidator( QString(":"), lin, "ascii_validator" ) );
      //	    if ( listview->propertyEditor()->formWindow()->isFake() )
      //		lin->setEnabled( FALSE );
    } else {
      lin->setValidator( new AsciiValidator( QString("!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~"
        "\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9"
        "\xaa\xab\xac\xad\xae\xaf\xb1\xb2\xb3"
        "\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc"
        "\xbd\xbe\xbf"), lin, "ascii_validator" ) );
    }
  } if ( !hasMultiLines ) {
    lin->hide();
  } else {
    button = new QPushButton( tr("..."), box );
    setupStyle( button );
    button->setFixedWidth( 20 );
    connect( button, SIGNAL( clicked() ),
      this, SLOT( getText() ) );
    lin->setFrame( FALSE );
  }
  connect( lin, SIGNAL( returnPressed() ),
    this, SLOT( setValue() ) );
  connect( lin, SIGNAL( lostFocus() ),
    this, SLOT( setValue() ) );
  //    connect( lin, SIGNAL( textChanged( const QString & ) ),
  //	     this, SLOT( setValue() ) );
  if ( PropertyItem::name() == "name" || PropertyItem::name() == "itemName" )
    //	connect( lin, SIGNAL( returnPressed() ),
    //		 listview->propertyEditor()->formWindow()->commandHistory(),
    //		 SLOT( checkCompressedCommand() ) );
    lin->installEventFilter( listview );
  return lin;
}

PropertyTextItem::~PropertyTextItem()
{
  delete (QLineEdit*)lin;
  lin = 0;
  delete (QHBox*)box;
  box = 0;
}

void PropertyTextItem::setChanged( bool b, bool updateDb )
{
  PropertyItem::setChanged( b, updateDb );
  if ( withComment && childCount() > 0 )
    ( (PropertyTextItem*)PropertyItem::child( 0 ) )->lined()->setEnabled( b );
}

bool PropertyTextItem::hasSubItems() const
{
  return withComment;
}

void PropertyTextItem::childValueChanged( PropertyItem * /* child */)
{
/*    if ( PropertyItem::name() != "name" )
MetaDataBase::setPropertyComment( listview->propertyEditor()->widget(),
PropertyItem::name(), child->value().toString() );
else
MetaDataBase::setExportMacro( listview->propertyEditor()->widget(), child->value().toString() );
listview->propertyEditor()->formWindow()->commandHistory()->setModified( TRUE );
  */
}

void PropertyTextItem::showEditor()
{
  if (!isEditable())
    return;
  PropertyItem::showEditor();
  if ( !lin || lin->text().length() == 0 ) {
    lined()->blockSignals( TRUE );
    lined()->setText( value().toString() );
    lined()->blockSignals( FALSE );
  }
  
  QWidget* w;
  if ( hasMultiLines )
    w = box;
  else
    w= lined();
  
  placeEditor( w );
  if ( !w->isVisible() || !lined()->hasFocus() ) {
    w->show();
    setFocus( lined() );
  }
}

void PropertyTextItem::createChildren()
{
  PropertyTextItem *i = new PropertyTextItem( listview, this, this,
    PropertyItem::name() == "name" ?
    "export macro" : "comment", FALSE, FALSE,
    PropertyItem::name() == "name" );
  i->lined()->setEnabled( isChanged() );
  addChild( i );
}

void PropertyTextItem::initChildren()
{
  if ( !childCount() )
    return;
  // PropertyItem *item = 
  PropertyItem::child( 0 );
  /*    if ( item ) {
  if ( PropertyItem::name() != "name" )
  item->setValue( MetaDataBase::propertyComment( listview->propertyEditor()->widget(),
  PropertyItem::name() ) );
  else
  item->setValue( MetaDataBase::exportMacro( listview->propertyEditor()->widget() ) );
  }
  */
}

void PropertyTextItem::hideEditor()
{
  if (!isEditable())
    return;
  PropertyItem::hideEditor();
  QWidget* w;
  if ( hasMultiLines )
    w = box;
  else
    w = lined();
  
  w->hide();
}

void PropertyTextItem::setValue( const QVariant &v )
{
  if ( ( !hasSubItems() || !isOpen() )
    && value() == v )
    return;
  if ( lin ) {
    lined()->blockSignals( TRUE );
    int oldCursorPos;
    oldCursorPos = lin->cursorPosition();
    lined()->setText( v.toString() );
    if ( oldCursorPos < (int)lin->text().length() )
      lin->setCursorPosition( oldCursorPos );
    lined()->blockSignals( FALSE );
  }
  setText( 1, v.toString() );
  PropertyItem::setValue( v );
}

void PropertyTextItem::setValue()
{
  setText( 1, lined()->text() );
  QVariant v;
  if ( accel ) {
    v = QVariant( QKeySequence( lined()->text() ) );
    if ( v.toString().isNull() )
      return; // not yet valid input
  } else {
    v = lined()->text();
  }
  PropertyItem::setValue( v );
  notifyValueChange();
}

void PropertyTextItem::getText()
{
//  bool richText = /* !::qt_cast<QButton*>(listview->propertyEditor()->widget()) || */
//    ( text( 0 ) == "whatsThis" );
//  bool doWrap = FALSE;
  QString txt = "";//MultiLineEditor::getText( listview, value().toString(), richText, &doWrap );
  if ( !txt.isEmpty() ) {
    setText( 1, txt );
    PropertyItem::setValue( txt );
    notifyValueChange();
    lined()->blockSignals( TRUE );
    lined()->setText( txt );
    lined()->blockSignals( FALSE );
  }
}

// --------------------------------------------------------------

PropertyDoubleItem::PropertyDoubleItem( PropertyList *l, PropertyItem *after, PropertyItem *prop,
                                       const QString &propName )
                                       : PropertyItem( l, after, prop, propName )
{
  lin = 0;
  _checkRange = false;
}

QLineEdit *PropertyDoubleItem::lined()
{
  if ( lin )
    return lin;
  lin = new QLineEdit( listview->viewport() );
  { // added by me
    lin->setText( QString::number( value().toDouble() ) );
  }
  if (_checkRange) {
    lin->setValidator( new QDoubleValidator( _bottom, _top, _decimals, 
      lin, "double_validator" ) );
  }
  else {
    lin->setValidator( new QDoubleValidator( lin, "double_validator" ) );
  }
  connect( lin, SIGNAL( returnPressed() ),
    this, SLOT( setValue() ) );
  connect( lin, SIGNAL( textChanged( const QString & ) ),
    this, SLOT( setValue() ) );
  lin->installEventFilter( listview );
  return lin;
}

PropertyDoubleItem::~PropertyDoubleItem()
{
  delete (QLineEdit*)lin;
  lin = 0;
}

void PropertyDoubleItem::showEditor()
{
  if (!isEditable())
    return;
  PropertyItem::showEditor();
  if ( !lin ) {
    lined()->blockSignals( TRUE );
    lined()->setText( QString::number( value().toDouble() ) );
    lined()->blockSignals( FALSE );
  }
  QWidget* w = lined();
  
  placeEditor( w );
  if ( !w->isVisible() || !lined()->hasFocus() ) {
    w->show();
    setFocus( lined() );
  }
}


void PropertyDoubleItem::hideEditor()
{
  if (!isEditable())
    return;
  PropertyItem::hideEditor();
  QWidget* w = lined();
  w->hide();
}

void PropertyDoubleItem::setValue( const QVariant &v )
{
  if ( value() == v )
    return;
  if ( lin ) {
    lined()->blockSignals( TRUE );
    int oldCursorPos;
    oldCursorPos = lin->cursorPosition();
    lined()->setText( QString::number( v.toDouble() ) );
    if ( oldCursorPos < (int)lin->text().length() )
      lin->setCursorPosition( oldCursorPos );
    lined()->blockSignals( FALSE );
  }
  setText( 1, QString::number( v.toDouble() ) );
  PropertyItem::setValue( v );
}

void PropertyDoubleItem::setValue()
{
  double vv = lined()->text().toDouble();
  QString s = lined()->text();
  if (_checkRange) {
    if (vv < _bottom) {
      vv = _bottom;
      s = QString::number(vv);
    }
    if (vv > _top) {
      vv = _top;
      s = QString::number(vv);
    }
  }
  setText( 1, s );
  if (s != lined()->text()) {
    lined()->blockSignals(true);
    lined()->setText(s);
    lined()->blockSignals(false);
  }
  QVariant v = vv;
  PropertyItem::setValue( v );
  notifyValueChange();
}


// --------------------------------------------------------------

PropertyDateItem::PropertyDateItem( PropertyList *l, PropertyItem *after, PropertyItem *prop, const QString &propName )
: PropertyItem( l, after, prop, propName )
{
  lin = 0;
}

QDateEdit *PropertyDateItem::lined()
{
  if ( lin )
    return lin;
  lin = new QDateEdit( listview->viewport() );
  QObjectList *l = lin->queryList( "QLineEdit" );
  for ( QObject *o = l->first(); o; o = l->next() )
    o->installEventFilter( listview );
  delete l;
  connect( lin, SIGNAL( valueChanged( const QDate & ) ),
    this, SLOT( setValue() ) );
  return lin;
}

PropertyDateItem::~PropertyDateItem()
{
  delete (QDateEdit*)lin;
  lin = 0;
}

void PropertyDateItem::showEditor()
{
  PropertyItem::showEditor();
  if ( !lin ) {
    lined()->blockSignals( TRUE );
    lined()->setDate( value().toDate() );
    lined()->blockSignals( FALSE );
  }
  placeEditor( lin );
  if ( !lin->isVisible() ) {
    lin->show();
    setFocus( lin );
  }
}

void PropertyDateItem::hideEditor()
{
  PropertyItem::hideEditor();
  if ( lin )
    lin->hide();
}

void PropertyDateItem::setValue( const QVariant &v )
{
  if ( ( !hasSubItems() || !isOpen() )
    && value() == v )
    return;
  
  if ( lin ) {
    lined()->blockSignals( TRUE );
    if ( lined()->date() != v.toDate() )
      lined()->setDate( v.toDate() );
    lined()->blockSignals( FALSE );
  }
  setText( 1, v.toDate().toString( ::Qt::ISODate ) );
  PropertyItem::setValue( v );
}

void PropertyDateItem::setValue()
{
  setText( 1, lined()->date().toString( ::Qt::ISODate ) );
  QVariant v;
  v = lined()->date();
  PropertyItem::setValue( v );
  notifyValueChange();
}

// --------------------------------------------------------------

PropertyTimeItem::PropertyTimeItem( PropertyList *l, PropertyItem *after, PropertyItem *prop, const QString &propName )
: PropertyItem( l, after, prop, propName )
{
  lin = 0;
}

QTimeEdit *PropertyTimeItem::lined()
{
  if ( lin )
    return lin;
  lin = new QTimeEdit( listview->viewport() );
  connect( lin, SIGNAL( valueChanged( const QTime & ) ),
    this, SLOT( setValue() ) );
  QObjectList *l = lin->queryList( "QLineEdit" );
  for ( QObject *o = l->first(); o; o = l->next() )
    o->installEventFilter( listview );
  delete l;
  return lin;
}

PropertyTimeItem::~PropertyTimeItem()
{
  delete (QTimeEdit*)lin;
  lin = 0;
}

void PropertyTimeItem::showEditor()
{
  PropertyItem::showEditor();
  if ( !lin ) {
    lined()->blockSignals( TRUE );
    lined()->setTime( value().toTime() );
    lined()->blockSignals( FALSE );
  }
  placeEditor( lin );
  if ( !lin->isVisible() ) {
    lin->show();
    setFocus( lin );
  }
}

void PropertyTimeItem::hideEditor()
{
  PropertyItem::hideEditor();
  if ( lin )
    lin->hide();
}

void PropertyTimeItem::setValue( const QVariant &v )
{
  if ( ( !hasSubItems() || !isOpen() )
    && value() == v )
    return;
  
  if ( lin ) {
    lined()->blockSignals( TRUE );
    if ( lined()->time() != v.toTime() )
      lined()->setTime( v.toTime() );
    lined()->blockSignals( FALSE );
  }
  setText( 1, v.toTime().toString( ::Qt::ISODate ) );
  PropertyItem::setValue( v );
}

void PropertyTimeItem::setValue()
{
  setText( 1, lined()->time().toString( ::Qt::ISODate ) );
  QVariant v;
  v = lined()->time();
  PropertyItem::setValue( v );
  notifyValueChange();
}

// --------------------------------------------------------------

PropertyDateTimeItem::PropertyDateTimeItem( PropertyList *l, PropertyItem *after, PropertyItem *prop, const QString &propName )
: PropertyItem( l, after, prop, propName )
{
  lin = 0;
}

QDateTimeEdit *PropertyDateTimeItem::lined()
{
  if ( lin )
    return lin;
  lin = new QDateTimeEdit( listview->viewport() );
  connect( lin, SIGNAL( valueChanged( const QDateTime & ) ),
    this, SLOT( setValue() ) );
  QObjectList *l = lin->queryList( "QLineEdit" );
  for ( QObject *o = l->first(); o; o = l->next() )
    o->installEventFilter( listview );
  delete l;
  return lin;
}

PropertyDateTimeItem::~PropertyDateTimeItem()
{
  delete (QDateTimeEdit*)lin;
  lin = 0;
}

void PropertyDateTimeItem::showEditor()
{
  PropertyItem::showEditor();
  if ( !lin ) {
    lined()->blockSignals( TRUE );
    lined()->setDateTime( value().toDateTime() );
    lined()->blockSignals( FALSE );
  }
  placeEditor( lin );
  if ( !lin->isVisible() ) {
    lin->show();
    setFocus( lin );
  }
}

void PropertyDateTimeItem::hideEditor()
{
  PropertyItem::hideEditor();
  if ( lin )
    lin->hide();
}

void PropertyDateTimeItem::setValue( const QVariant &v )
{
  if ( ( !hasSubItems() || !isOpen() )
    && value() == v )
    return;
  
  if ( lin ) {
    lined()->blockSignals( TRUE );
    if ( lined()->dateTime() != v.toDateTime() )
      lined()->setDateTime( v.toDateTime() );
    lined()->blockSignals( FALSE );
  }
  setText( 1, v.toDateTime().toString( ::Qt::ISODate ) );
  PropertyItem::setValue( v );
}

void PropertyDateTimeItem::setValue()
{
  setText( 1, lined()->dateTime().toString( ::Qt::ISODate ) );
  QVariant v;
  v = lined()->dateTime();
  PropertyItem::setValue( v );
  notifyValueChange();
}

// --------------------------------------------------------------

PropertyBoolItem::PropertyBoolItem( PropertyList *l, PropertyItem *after, PropertyItem *prop, const QString &propName )
: PropertyItem( l, after, prop, propName )
{
  comb = 0;
}

QComboBox *PropertyBoolItem::combo()
{
  if ( comb )
    return comb;
  comb = new QComboBox( FALSE, listview->viewport() );
  comb->hide();
  comb->insertItem( tr( "False" ) );
  comb->insertItem( tr( "True" ) );
  connect( comb, SIGNAL( activated( int ) ),
    this, SLOT( setValue() ) );
  comb->installEventFilter( listview );
  return comb;
}

PropertyBoolItem::~PropertyBoolItem()
{
  delete (QComboBox*)comb;
  comb = 0;
}

void PropertyBoolItem::toggle()
{
  bool b = value().toBool();
  setValue( QVariant( !b, 0 ) );
  setValue();
}

void PropertyBoolItem::showEditor()
{
  PropertyItem::showEditor();
  if ( !comb ) {
    combo()->blockSignals( TRUE );
    if ( value().toBool() )
      combo()->setCurrentItem( 1 );
    else
      combo()->setCurrentItem( 0 );
    combo()->blockSignals( FALSE );
  }
  placeEditor( combo() );
  if ( !combo()->isVisible()  || !combo()->hasFocus() ) {
    combo()->show();
    setFocus( combo() );
  }
}

void PropertyBoolItem::hideEditor()
{
  PropertyItem::hideEditor();
  combo()->hide();
}

void PropertyBoolItem::setValue( const QVariant &v )
{
  if ( ( !hasSubItems() || !isOpen() )
    && value() == v )
    return;
  
  if ( comb ) {
    combo()->blockSignals( TRUE );
    if ( v.toBool() )
      combo()->setCurrentItem( 1 );
    else
      combo()->setCurrentItem( 0 );
    combo()->blockSignals( FALSE );
  }
  QString tmp = tr( "True" );
  if ( !v.toBool() )
    tmp = tr( "False" );
  setText( 1, tmp );
  PropertyItem::setValue( v );
}

void PropertyBoolItem::setValue()
{
  if ( !comb )
    return;
  setText( 1, combo()->currentText() );
  bool b = combo()->currentItem() == 0 ? (bool)FALSE : (bool)TRUE;
  PropertyItem::setValue( QVariant( b, 0 ) );
  notifyValueChange();
}

// --------------------------------------------------------------

PropertyIntItem::PropertyIntItem( PropertyList *l, PropertyItem *after, PropertyItem *prop,
                                 const QString &propName, bool s, int* lowBound, int* upperBound )
                                 : PropertyItem( l, after, prop, propName ), signedValue( s )
{
  spinBx = 0;
  if (lowBound != 0) {
    _lowBound = *lowBound;
  }
  else {
    if (s) {
      _lowBound = -INT_MAX;
    }
    else {
      _lowBound = 0;
    }
  }
  if (upperBound != 0) {
    _upperBound = *upperBound;
  }
  else {
    _upperBound = INT_MAX;
  }
}

QSpinBox *PropertyIntItem::spinBox()
{
  if ( spinBx )
    return spinBx;
  spinBx = new QSpinBox( _lowBound, _upperBound, 1, listview->viewport() );
  spinBx->setValue( value().toInt() );
  spinBx->hide();
  spinBx->installEventFilter( listview );
  QObjectList *ol = spinBx->queryList( "QLineEdit" );
  if ( ol && ol->first() )
    ol->first()->installEventFilter( listview );
  delete ol;
  connect( spinBx, SIGNAL( valueChanged( int ) ),
	   this, SLOT( setValue() ) );
  return spinBx;
}

PropertyIntItem::~PropertyIntItem()
{
  delete (QSpinBox*)spinBx;
  spinBx = 0;
}

void PropertyIntItem::showEditor()
{
  if (!isEditable())
    return;
  PropertyItem::showEditor();
  if ( !spinBx ) {
    spinBox()->blockSignals( TRUE );
    if ( signedValue )
      spinBox()->setValue( value().toInt() );
    else
      spinBox()->setValue( value().toUInt() );
    spinBox()->blockSignals( FALSE );
  }
  placeEditor( spinBox() );
  if ( !spinBox()->isVisible()  || !spinBox()->hasFocus()  ) {
    spinBox()->show();
    setFocus( spinBox() );
  }
}

void PropertyIntItem::hideEditor()
{
  if (!isEditable())
    return;
  PropertyItem::hideEditor();
  spinBox()->hide();
}

void PropertyIntItem::setValue( const QVariant &v )
{
  if ( ( !hasSubItems() || !isOpen() )
    && value() == v )
    return;
  
  if ( spinBx ) {
    spinBox()->blockSignals( TRUE );
    if ( signedValue )
      spinBox()->setValue( v.toInt() );
    else
      spinBox()->setValue( v.toUInt() );
    spinBox()->blockSignals( FALSE );
  }
  
  if ( signedValue )
    setText( 1, QString::number( v.toInt() ) );
  else
    setText( 1, QString::number( v.toUInt() ) );
  PropertyItem::setValue( v );
}

void PropertyIntItem::setValue()
{
  if ( !spinBx )
    return;
  setText( 1, QString::number( spinBox()->value() ) );
  if ( signedValue )
    PropertyItem::setValue( spinBox()->value() );
  else
    PropertyItem::setValue( (uint)spinBox()->value() );
  notifyValueChange();
}



// --------------------------------------------------------------

PropertyListItem::PropertyListItem( PropertyList *l, PropertyItem *after, PropertyItem *prop,
                                   const QString &propName, bool e )
                                   : PropertyItem( l, after, prop, propName ), editable( e )
{
  comb = 0;
  oldInt = -1;
}

QComboBox *PropertyListItem::combo()
{
  if ( comb )
    return comb;
  comb = new QComboBox( editable, listview->viewport() );
  comb->hide();
  connect( comb, SIGNAL( activated( int ) ),
    this, SLOT( setValue() ) );
  comb->installEventFilter( listview );
  if ( editable ) {
    QObjectList *ol = comb->queryList( "QLineEdit" );
    if ( ol && ol->first() )
      ol->first()->installEventFilter( listview );
    delete ol;
  }
  return comb;
}

PropertyListItem::~PropertyListItem()
{
  delete (QComboBox*)comb;
  comb = 0;
}

void PropertyListItem::showEditor()
{
  if (!isEditable())
    return;
  PropertyItem::showEditor();
  if ( !comb ) {
    combo()->blockSignals( TRUE );
    combo()->clear();
    combo()->insertStringList( value().toStringList() );
    combo()->blockSignals( FALSE );
  }
  placeEditor( combo() );
  if ( !combo()->isVisible()  || !combo()->hasFocus() ) {
    combo()->show();
    setFocus( combo() );
  }
}

void PropertyListItem::hideEditor()
{
  if (!isEditable())
    return;
  PropertyItem::hideEditor();
  combo()->hide();
}

void PropertyListItem::setValue( const QVariant &v )
{
  if ( comb ) {
    combo()->blockSignals( TRUE );
    combo()->clear();
    combo()->insertStringList( v.toStringList() );
    combo()->blockSignals( FALSE );
  }
  setText( 1, v.toStringList().first() );
  PropertyItem::setValue( v );
}

void PropertyListItem::setValue()
{
  if ( !comb )
    return;
  setText( 1, combo()->currentText() );
  QStringList lst;
  for ( uint i = 0; i < combo()->listBox()->count(); ++i )
    lst << combo()->listBox()->item( i )->text();
  PropertyItem::setValue( lst );
  notifyValueChange();
  oldInt = currentIntItem();
  oldString = currentItem();
}

QString PropertyListItem::currentItem() const
{
  return ( (PropertyListItem*)this )->combo()->currentText();
}

void PropertyListItem::setCurrentItem( const QString &s )
{
  if ( comb && currentItem().lower() == s.lower() )
    return;
  
  if ( !comb ) {
    combo()->blockSignals( TRUE );
    combo()->clear();
    combo()->insertStringList( value().toStringList() );
    combo()->blockSignals( FALSE );
  }
  for ( uint i = 0; i < combo()->listBox()->count(); ++i ) {
    if ( combo()->listBox()->item( i )->text().lower() == s.lower() ) {
      combo()->setCurrentItem( i );
      setText( 1, combo()->currentText() );
      break;
    }
  }
  oldInt = currentIntItem();
  oldString = currentItem();
}

void PropertyListItem::addItem( const QString &s )
{
  combo()->insertItem( s );
}

void PropertyListItem::setCurrentItem( int i )
{
  if ( comb && i == combo()->currentItem() )
    return;
  
  if ( !comb ) {
    combo()->blockSignals( TRUE );
    combo()->clear();
    combo()->insertStringList( value().toStringList() );
    combo()->blockSignals( FALSE );
  }
  combo()->setCurrentItem( i );
  setText( 1, combo()->currentText() );
  oldInt = currentIntItem();
  oldString = currentItem();
}

int PropertyListItem::currentIntItem() const
{
  return ( (PropertyListItem*)this )->combo()->currentItem();
}

int PropertyListItem::currentIntItemFromObject() const
{
  return oldInt;
}

QString PropertyListItem::currentItemFromObject() const
{
  return oldString;
}

// --------------------------------------------------------------

PropertyPixmapItem::PropertyPixmapItem( PropertyList *l, PropertyItem *after, PropertyItem *prop,
                                       const QString &propName, Type t )
                                       : PropertyItem( l, after, prop, propName ), type( t )
{
  box = new QHBox( listview->viewport() );
  box->hide();
  pixPrev = new QLabel( box );
  pixPrev->setSizePolicy( QSizePolicy( QSizePolicy::Ignored, QSizePolicy::Minimum ) );
  pixPrev->setBackgroundColor( pixPrev->colorGroup().color( QColorGroup::Base ) );
  button = new QPushButton( "...", box );
  setupStyle( button );
  button->setFixedWidth( 20 );
  box->setFrameStyle( QFrame::StyledPanel | QFrame::Sunken );
  box->setLineWidth( 2 );
  pixPrev->setFrameStyle( QFrame::NoFrame );
  box->installEventFilter( listview );
  connect( button, SIGNAL( clicked() ),
    this, SLOT( getPixmap() ) );
}

PropertyPixmapItem::~PropertyPixmapItem()
{
  delete (QHBox*)box;
}

void PropertyPixmapItem::showEditor()
{
  PropertyItem::showEditor();
  placeEditor( box );
  if ( !box->isVisible() ) {
    box->show();
    listView()->viewport()->setFocus();
  }
}

void PropertyPixmapItem::hideEditor()
{
  PropertyItem::hideEditor();
  box->hide();
}

void PropertyPixmapItem::setValue( const QVariant &v )
{
  QString s;
  if ( type == Pixmap )
    pixPrev->setPixmap( v.toPixmap() );
  else if ( type == IconSet )
    pixPrev->setPixmap( v.toIconSet().pixmap() );
  else
    pixPrev->setPixmap( v.toImage() );
  PropertyItem::setValue( v );
  repaint();
}

void PropertyPixmapItem::getPixmap()
{
  QPixmap pix;// = qChoosePixmap( listview, listview->propertyEditor()->formWindow(), value().toPixmap() );
  if ( !pix.isNull() ) {
    if ( type == Pixmap )
      setValue( pix );
    else if ( type == IconSet )
      setValue( QIconSet( pix ) );
    else
      setValue( pix.convertToImage() );
    
    notifyValueChange();
  }
}

bool PropertyPixmapItem::hasCustomContents() const
{
  return TRUE;
}

void PropertyPixmapItem::drawCustomContents( QPainter *p, const QRect &r )
{
  QPixmap pix;
  if ( type == Pixmap )
    pix = value().toPixmap();
  else if ( type == IconSet )
    pix = value().toIconSet().pixmap();
  else
    pix = value().toImage();
  
  if ( !pix.isNull() ) {
    p->save();
    p->setClipRect( QRect( QPoint( (int)(p->worldMatrix().dx() + r.x()),
      (int)(p->worldMatrix().dy() + r.y()) ),
      r.size() ) );
    p->drawPixmap( r.x(), r.y() + ( r.height() - pix.height() ) / 2, pix );
    p->restore();
  }
}


// --------------------------------------------------------------

PropertyColorItem::PropertyColorItem( PropertyList *l, PropertyItem *after, PropertyItem *prop,
                                     const QString &propName, bool children )
                                     : PropertyItem( l, after, prop, propName ), withChildren( children )
{
  box = new QHBox( listview->viewport() );
  box->hide();
  colorPrev = new QFrame( box );
  button = new QPushButton( "...", box );
  setupStyle( button );
  button->setFixedWidth( 20 );
  box->setFrameStyle( QFrame::StyledPanel | QFrame::Sunken );
  box->setLineWidth( 2 );
  colorPrev->setFrameStyle( QFrame::Plain | QFrame::Box );
  colorPrev->setLineWidth( 2 );
  QPalette pal = colorPrev->palette();
  QColorGroup cg = pal.active();
  cg.setColor( QColorGroup::Foreground, cg.color( QColorGroup::Base ) );
  pal.setActive( cg );
  pal.setInactive( cg );
  pal.setDisabled( cg );
  colorPrev->setPalette( pal );
  box->installEventFilter( listview );
  connect( button, SIGNAL( clicked() ),
    this, SLOT( getColor() ) );
}

void PropertyColorItem::createChildren()
{
  PropertyItem *i = this;
  i = new PropertyIntItem( listview, i, this, tr( "Red" ), TRUE );
  addChild( i );
  i = new PropertyIntItem( listview, i, this, tr( "Green" ), TRUE );
  addChild( i );
  i = new PropertyIntItem( listview, i, this, tr( "Blue" ), TRUE );
  addChild( i );
}

void PropertyColorItem::initChildren()
{
  PropertyItem *item = 0;
  for ( int i = 0; i < childCount(); ++i ) {
    item = PropertyItem::child( i );
    if ( item->name() == tr( "Red" ) )
      item->setValue( val.toColor().red() );
    else if ( item->name() == tr( "Green" ) )
      item->setValue( val.toColor().green() );
    else if ( item->name() == tr( "Blue" ) )
      item->setValue( val.toColor().blue() );
  }
}

PropertyColorItem::~PropertyColorItem()
{
  delete (QHBox*)box;
}

void PropertyColorItem::showEditor()
{
  PropertyItem::showEditor();
  placeEditor( box );
  if ( !box->isVisible() ) {
    box->show();
    listView()->viewport()->setFocus();
  }
}

void PropertyColorItem::hideEditor()
{
  PropertyItem::hideEditor();
  box->hide();
}

void PropertyColorItem::setValue( const QVariant &v )
{
  if ( ( !hasSubItems() || !isOpen() )
    && value() == v )
    return;
  
  QString s;
  setText( 1, v.toColor().name() );
  colorPrev->setBackgroundColor( v.toColor() );
  PropertyItem::setValue( v );
}

bool PropertyColorItem::hasSubItems() const
{
  return withChildren;
}

void PropertyColorItem::childValueChanged( PropertyItem *child )
{
  QColor c( val.toColor() );
  if ( child->name() == tr( "Red" ) )
    c.setRgb( child->value().toInt(), c.green(), c.blue() );
  else if ( child->name() == tr( "Green" ) )
    c.setRgb( c.red(), child->value().toInt(), c.blue() );
  else if ( child->name() == tr( "Blue" ) )
    c.setRgb( c.red(), c.green(), child->value().toInt() );
  setValue( c );
  notifyValueChange();
}

void PropertyColorItem::getColor()
{
  QColor c = QColorDialog::getColor( val.asColor(), listview );
  if ( c.isValid() ) {
    setValue( c );
    notifyValueChange();
  }
}

bool PropertyColorItem::hasCustomContents() const
{
  return TRUE;
}

void PropertyColorItem::drawCustomContents( QPainter *p, const QRect &r )
{
  p->save();
  p->setPen( QPen( black, 1 ) );
  p->setBrush( val.toColor() );
  p->drawRect( r.x() + 2, r.y() + 2, r.width() - 5, r.height() - 5 );
  p->restore();
}

// --------------------------------------------------------------

PropertyFontItem::PropertyFontItem( PropertyList *l, PropertyItem *after, PropertyItem *prop, const QString &propName )
: PropertyItem( l, after, prop, propName )
{
  box = new QHBox( listview->viewport() );
  box->hide();
  lined = new QLineEdit( box );
  button = new QPushButton( "...", box );
  setupStyle( button );
  button->setFixedWidth( 20 );
  box->setFrameStyle( QFrame::StyledPanel | QFrame::Sunken );
  box->setLineWidth( 2 );
  lined->setFrame( FALSE );
  lined->setReadOnly( TRUE );
  box->setFocusProxy( lined );
  box->installEventFilter( listview );
  lined->installEventFilter( listview );
  button->installEventFilter( listview );
  connect( button, SIGNAL( clicked() ),
    this, SLOT( getFont() ) );
}

void PropertyFontItem::createChildren()
{
  PropertyItem *i = this;
  i = new PropertyListItem( listview, i, this, tr( "Family" ), FALSE );
  addChild( i );
  i = new PropertyIntItem( listview, i, this, tr( "Point Size" ), TRUE );
  addChild( i );
  i = new PropertyBoolItem( listview, i, this, tr( "Bold" ) );
  addChild( i );
  i = new PropertyBoolItem( listview, i, this, tr( "Italic" ) );
  addChild( i );
  i = new PropertyBoolItem( listview, i, this, tr( "Underline" ) );
  addChild( i );
  i = new PropertyBoolItem( listview, i, this, tr( "Strikeout" ) );
  addChild( i );
}

void PropertyFontItem::initChildren()
{
  PropertyItem *item = 0;
  for ( int i = 0; i < childCount(); ++i ) {
    item = PropertyItem::child( i );
    if ( item->name() == tr( "Family" ) ) {
      ( (PropertyListItem*)item )->setValue( getFontList() );
      ( (PropertyListItem*)item )->setCurrentItem( val.toFont().family() );
    } else if ( item->name() == tr( "Point Size" ) )
      item->setValue( val.toFont().pointSize() );
    else if ( item->name() == tr( "Bold" ) )
      item->setValue( QVariant( val.toFont().bold(), 0 ) );
    else if ( item->name() == tr( "Italic" ) )
      item->setValue( QVariant( val.toFont().italic(), 0 ) );
    else if ( item->name() == tr( "Underline" ) )
      item->setValue( QVariant( val.toFont().underline(), 0 ) );
    else if ( item->name() == tr( "Strikeout" ) )
      item->setValue( QVariant( val.toFont().strikeOut(), 0 ) );
  }
}

PropertyFontItem::~PropertyFontItem()
{
  delete (QHBox*)box;
}

void PropertyFontItem::showEditor()
{
  PropertyItem::showEditor();
  placeEditor( box );
  if ( !box->isVisible() || !lined->hasFocus() ) {
    box->show();
    setFocus( lined );
  }
}

void PropertyFontItem::hideEditor()
{
  PropertyItem::hideEditor();
  box->hide();
}

void PropertyFontItem::setValue( const QVariant &v )
{
  if ( value() == v )
    return;
  
  setText( 1, v.toFont().family() + "-" + QString::number( v.toFont().pointSize() ) );
  lined->setText( v.toFont().family() + "-" + QString::number( v.toFont().pointSize() ) );
  PropertyItem::setValue( v );
}

void PropertyFontItem::getFont()
{
  bool ok = FALSE;
  QFont f = QFontDialog::getFont( &ok, val.toFont(), listview );
  if ( ok && f != val.toFont() ) {
    setValue( f );
    notifyValueChange();
  }
}

bool PropertyFontItem::hasSubItems() const
{
  return TRUE;
}

void PropertyFontItem::childValueChanged( PropertyItem *child )
{
  QFont f = val.toFont();
  if ( child->name() == tr( "Family" ) )
    f.setFamily( ( (PropertyListItem*)child )->currentItem() );
  else if ( child->name() == tr( "Point Size" ) )
    f.setPointSize( child->value().toInt() );
  else if ( child->name() == tr( "Bold" ) )
    f.setBold( child->value().toBool() );
  else if ( child->name() == tr( "Italic" ) )
    f.setItalic( child->value().toBool() );
  else if ( child->name() == tr( "Underline" ) )
    f.setUnderline( child->value().toBool() );
  else if ( child->name() == tr( "Strikeout" ) )
    f.setStrikeOut( child->value().toBool() );
  setValue( f );
  notifyValueChange();
}


// --------------------------------------------------------------

EnumPopup::EnumPopup( QWidget *parent, const char *name, WFlags f )
: QFrame( parent, name, f )
{
  setLineWidth( 1 );
  setFrameStyle( Panel | Plain );
  setPaletteBackgroundColor( Qt::white );
  popLayout = new QVBoxLayout( this, 3 );
  checkBoxList.setAutoDelete( TRUE );
}

EnumPopup::~EnumPopup()
{
}

void EnumPopup::insertEnums( QValueList<EnumItem> lst )
{
  while ( checkBoxList.count() )
    checkBoxList.removeFirst();
  
  itemList = lst;
  QCheckBox *cb;
  QValueListConstIterator<EnumItem> it = itemList.begin();
  for ( ; it != itemList.end(); ++it ) {
    cb = new QCheckBox( this );
    cb->setText( (*it).key );
    cb->setChecked( (*it).selected );
    if ( it == itemList.begin() )
      cb->setFocus();
    checkBoxList.append( cb );
    cb->resize( width(), cb->height() );
    popLayout->addWidget( cb );
  }
}

void EnumPopup::keyPressEvent( QKeyEvent *e )
{
  if ( e->key() == Key_Escape ) {
    hide();
    emit hidden();
  } else if ( e->key() == Key_Enter || e->key() == Key_Return ) {
    closeWidget();
  }
}

void EnumPopup::closeWidget()
{
  QPtrListIterator<QCheckBox> it( checkBoxList );
  int i = 0;
  while ( it.current() != 0 ) {
    itemList[i].selected = (*it)->isChecked();
    ++it;
    ++i;
  }
  close();
  emit closed();
}

QValueList<EnumItem> EnumPopup::enumList() const
{
  return itemList;
}

EnumBox::EnumBox( QWidget *parent, const char *name )
: QComboBox( parent, name )
{
  pop = new EnumPopup( this, "popup", QObject::WType_Popup );
  connect( pop, SIGNAL( hidden() ), this, SLOT( popupHidden() ) );
  connect( pop, SIGNAL( closed() ), this, SLOT( popupClosed() ) );
  popupShown = FALSE;
  arrowDown = FALSE;
}

void EnumBox::popupHidden()
{
  popupShown = FALSE;
}

void EnumBox::popupClosed()
{
  popupShown = FALSE;
  emit valueChanged();
}

void EnumBox::paintEvent( QPaintEvent * )
{
  QPainter p( this );
  const QColorGroup & g = colorGroup();
  p.setPen(g.text());
  
  QStyle::SFlags flags = QStyle::Style_Default;
  if (isEnabled())
    flags |= QStyle::Style_Enabled;
  if (hasFocus())
    flags |= QStyle::Style_HasFocus;
  
  if ( width() < 5 || height() < 5 ) {
    qDrawShadePanel( &p, rect().x(), rect().y(), rect().width(), rect().height(), g, FALSE, 2,
      &g.brush( QColorGroup::Button ) );
    return;
  }
  style().drawComplexControl( QStyle::CC_ComboBox, &p, this, rect(), g,
    flags, QStyle::SC_All,
    (arrowDown ?
    QStyle::SC_ComboBoxArrow :
  QStyle::SC_None ));
  
  QRect re = style().querySubControlMetrics( QStyle::CC_ComboBox, this,
    QStyle::SC_ComboBoxEditField );
  re = QStyle::visualRect(re, this);
  p.setClipRect( re );
  
  if ( !str.isNull() ) {
    p.save();
    p.setFont(font());
    QFontMetrics fm(font());
    int x = re.x(), y = re.y() + fm.ascent();
    p.drawText( x, y, str );
    p.restore();
  }
}

void EnumBox::insertEnums( QValueList<EnumItem> lst )
{
  pop->insertEnums( lst );
}

QValueList<EnumItem> EnumBox::enumList() const
{
  return pop->enumList();
}

void EnumBox::popup()
{
  if ( popupShown ) {
    pop->closeWidget();
    popupShown = FALSE;
    return;
  }
  pop->move( ((QWidget*)parent())->mapToGlobal( geometry().bottomLeft() ) );
  pop->setMinimumWidth( width() );
  emit aboutToShowPopup();
  pop->show();
  popupShown = TRUE;
}

void EnumBox::mousePressEvent( QMouseEvent *e )
{
  if ( e->button() != LeftButton )
    return;
  
  QRect arrowRect = style().querySubControlMetrics( QStyle::CC_ComboBox, this,
    QStyle::SC_ComboBoxArrow);
  arrowRect = QStyle::visualRect(arrowRect, this);
  
  arrowRect.setHeight( QMAX(  height() - (2 * arrowRect.y()), arrowRect.height() ) );
  
  if ( arrowRect.contains( e->pos() ) ) {
    arrowDown = TRUE;
    repaint( FALSE );
  }
  
  popup();
  QTimer::singleShot( 100, this, SLOT( restoreArrow() ) );
}

void EnumBox::keyPressEvent( QKeyEvent *e )
{
  if ( e->key() == Key_Space ) {
    popup();
    QTimer::singleShot( 100, this, SLOT( restoreArrow() ) );
  } else if ( e->key() == Key_Enter || e->key() == Key_Return ) {
    popup();
  }
}

void EnumBox::restoreArrow()
{
  arrowDown = FALSE;
  repaint( FALSE );
}

void EnumBox::setText( const QString &text )
{
  str = text;
  repaint( FALSE );
}



PropertyEnumItem::PropertyEnumItem( PropertyList *l,
                                   PropertyItem *after,
                                   PropertyItem *prop,
                                   const QString &propName )
                                   : PropertyItem( l, after, prop, propName )
{
  box = new EnumBox( listview->viewport() );
  box->hide();
  box->installEventFilter( listview );
  connect( box, SIGNAL( aboutToShowPopup() ), this, SLOT( insertEnums() ) );
  connect( box, SIGNAL( valueChanged() ), this, SLOT( setValue() ) );
}

PropertyEnumItem::~PropertyEnumItem()
{
  delete (EnumBox*)box;
}

void PropertyEnumItem::showEditor()
{
  PropertyItem::showEditor();
  placeEditor( box );
  if ( !box->isVisible() ) {
    box->show();
    box->setText( enumString );
    listView()->viewport()->setFocus();
  }
  box->setFocus();
}

void PropertyEnumItem::hideEditor()
{
  PropertyItem::hideEditor();
  box->hide();
}

void PropertyEnumItem::setValue( const QVariant &v )
{
  enumString = "";
  enumList.clear();
  QStringList lst = v.toStringList();
  QValueListConstIterator<QString> it = lst.begin();
  for ( ; it != lst.end(); ++it )
    enumList.append( EnumItem( *it, FALSE ) );
  enumList.first().selected = TRUE;
  enumString = enumList.first().key;
  box->setText( enumString );
  setText( 1, enumString );
  PropertyItem::setValue( v );
}

void PropertyEnumItem::insertEnums()
{
  box->insertEnums( enumList );
}

void PropertyEnumItem::setValue()
{
  enumList = box->enumList();
  enumString = "";
  QValueListConstIterator<EnumItem> it = enumList.begin();
  for ( ; it != enumList.end(); ++it ) {
    if ( (*it).selected )
      enumString += "|" + (*it).key;
  }
  if ( !enumString.isEmpty() )
    enumString.replace( 0, 1, "" );
  
  box->setText( enumString );
  setText( 1, enumString );
  notifyValueChange();
}

void PropertyEnumItem::setCurrentValues( QStrList lst )
{
  enumString ="";
  QStrList::ConstIterator it = lst.begin();
  QValueList<EnumItem>::Iterator eit = enumList.begin();
  for ( ; eit != enumList.end(); ++eit ) {
    (*eit).selected = FALSE;
    for ( it = lst.begin(); it != lst.end(); ++it ) {
      if ( QString( *it ) == (*eit).key ) {
        (*eit).selected = TRUE;
        enumString += "|" + (*eit).key;
        break;
      }
    }
  }
  if ( !enumString.isEmpty() )
    enumString.replace( 0, 1, "" );
  box->setText( enumString );
  setText( 1, enumString );
}

QString PropertyEnumItem::currentItem() const
{
  return enumString;
}

QString PropertyEnumItem::currentItemFromObject() const
{
  return enumString;
}

// --------------------------------------------------------------

/*!
\class PropertyList propertyeditor.h
\brief PropertyList is a QListView derived class which is used for editing widget properties

  This class is used for widget properties. It has to be child of a
  PropertyEditor.
  
    To initialize it for editing a widget call setupProperties() which
    iterates through the properties of the current widget (see
    PropertyEditor::widget()) and builds the list.
    
      To update the item values, refetchData() can be called.
      
        If the value of an item has been changed by the user, and this
        change should be applied to the widget's property, valueChanged()
        has to be called.
        
          To set the value of an item, setPropertyValue() has to be called.
*/

PropertyList::PropertyList( QWidget* parent, const char* name  )
: QListView( parent, name)
{
  init_colors();
  
  whatsThis = new PropertyWhatsThis( this );
  showSorted = FALSE;
  header()->setMovingEnabled( FALSE );
  header()->setStretchEnabled( TRUE );
  setResizePolicy( QScrollView::Manual );
  viewport()->setAcceptDrops( TRUE );
  viewport()->installEventFilter( this );
  addColumn( tr( "Property" ) );
  addColumn( tr( "Value" ) );
  connect( header(), SIGNAL( sizeChange( int, int, int ) ),
    this, SLOT( updateEditorSize() ) );
  disconnect( header(), SIGNAL( sectionClicked( int ) ),
    this, SLOT( changeSortColumn( int ) ) );
  connect( header(), SIGNAL( sectionClicked( int ) ),
    this, SLOT( toggleSort() ) );
  connect( this, SIGNAL( pressed( QListViewItem *, const QPoint &, int ) ),
    this, SLOT( itemPressed( QListViewItem *, const QPoint &, int ) ) );
  connect( this, SIGNAL( doubleClicked( QListViewItem * ) ),
    this, SLOT( toggleOpen( QListViewItem * ) ) );
  setSorting( -1 );
  setHScrollBarMode( AlwaysOff );
  setVScrollBarMode( AlwaysOn );
  setColumnWidthMode( 1, Manual );
  mousePressed = FALSE;
  pressItem = 0;
  theLastEvent = MouseEvent;
  header()->installEventFilter( this );
}

PropertyList::~PropertyList() {
  clear();
}

void PropertyList::toggleSort()
{
  showSorted = !showSorted;
  //    editor->clear();
  //    editor->setup();
}

void PropertyList::resizeEvent( QResizeEvent *e )
{
  QListView::resizeEvent( e );
  if ( currentItem() )
    ( ( PropertyItem* )currentItem() )->showEditor();
}

static QVariant::Type type_to_variant( const QString &s )
{
  if ( s == "Invalid " )
    return QVariant::Invalid;
  if ( s == "Map" )
    return QVariant::Map;
  if ( s == "List" )
    return QVariant::List;
  if ( s == "String" )
    return QVariant::String;
  if ( s == "StringList" )
    return QVariant::StringList;
  if ( s == "Font" )
    return QVariant::Font;
  if ( s == "Pixmap" )
    return QVariant::Pixmap;
  if ( s == "Brush" )
    return QVariant::Brush;
  if ( s == "Rect" )
    return QVariant::Rect;
  if ( s == "Size" )
    return QVariant::Size;
  if ( s == "Color" )
    return QVariant::Color;
  if ( s == "Palette" )
    return QVariant::Palette;
  if ( s == "ColorGroup" )
    return QVariant::ColorGroup;
  if ( s == "IconSet" )
    return QVariant::IconSet;
  if ( s == "Point" )
    return QVariant::Point;
  if ( s == "Image" )
    return QVariant::Image;
  if ( s == "Int" )
    return QVariant::Int;
  if ( s == "UInt" )
    return QVariant::UInt;
  if ( s == "Bool" )
    return QVariant::Bool;
  if ( s == "Double" )
    return QVariant::Double;
  if ( s == "CString" )
    return QVariant::CString;
  if ( s == "PointArray" )
    return QVariant::PointArray;
  if ( s == "Region" )
    return QVariant::Region;
  if ( s == "Bitmap" )
    return QVariant::Bitmap;
  if ( s == "Cursor" )
    return QVariant::Cursor;
  if ( s == "SizePolicy" )
    return QVariant::SizePolicy;
  if ( s == "Date" )
    return QVariant::Date;
  if ( s == "Time" )
    return QVariant::Time;
  if ( s == "DateTime" )
    return QVariant::DateTime;
  return QVariant::Invalid;
}

/*#ifndef QT_NO_SQL
static bool parent_is_data_aware( QWidget *w )
{
QWidget *p = w ? w->parentWidget() : 0;
while ( p && !p->isTopLevel() ) {
if ( ::qt_cast<QDesignerDataBrowser*>(p) || ::qt_cast<QDesignerDataView*>(p) )
return TRUE;
p = p->parentWidget();
}
return FALSE;
}
#endif
*/
/*!  Sets up the property list by adding an item for each designable
property of the widget which is just edited.
*/

void PropertyList::setupProperties(const BProperties& props)
{
  blockSignals(true);
  clear();
  PropertyItem *item = 0;
  for(BProperties::const_iterator cit = props.begin(); cit != props.end(); cit++) {
    BProperty* p = new BProperty(**cit);
    _properties.addProperty(p);
    if ( !addPropertyItem( item, p ) )
      continue;
    item->setValue(p->value());
    if (p->type() == BProperty::StringListType) {
      item->setCurrentItem(p->options().toInt());
    }
    _items[item] = p;
  }      
  setCurrentItem( firstChild() );
  
  if ( showSorted ) {
    setSorting( 0 );
    sort();
    setSorting( -1 );
    setCurrentItem( firstChild() );
    qApp->processEvents();
  }
  
  updateEditorSize();
  blockSignals(false);
}

void PropertyList::updateProperties(const BProperties& props) 
{
  blockSignals(true);
  if (props.size() != _properties.size()) {
    throw BInternalException(__FILE__, __LINE__,
      tr("Requested '%1' properties but only '%2' available").arg(
      props.size()).arg(_properties.size()));
  }
  for(BProperties::const_iterator cit = props.begin(); 
  cit != props.end(); cit++) {
    bool propFound = false;
    for (BPropertyItems::iterator it = _items.begin(); 
    it != _items.end(); it++) {
      if (it->second->key() == (*cit)->key()) {
        propFound = true;
        if (it->first->value() != (*cit)->value())
          it->first->setValue((*cit)->value());
        if ((*cit)->type() == BProperty::StringListType) {
          it->first->setCurrentItem((*cit)->options().toInt());
        }
        break;
      }
    }
    if (!propFound) {
      throw BInternalException(__FILE__, __LINE__,
        tr("Property '%1' reguested but not found").arg((*cit)->key()));
    }
  }
//  qDebug("update ship view");
  blockSignals(false);
}

void PropertyList::getProperties(BProperties& props) 
{
  if (props.size() != _properties.size()) {
    throw BInternalException(__FILE__, __LINE__,
      tr("Requested '%1' properties but only '%2' available").arg(
      props.size()).arg(_properties.size()));
  }
  for(BProperties::iterator cit = props.begin(); 
  cit != props.end(); cit++) {
    bool propFound = false;
    for (BPropertyItems::iterator it = _items.begin(); 
    it != _items.end(); it++) {
      if (it->second->key() == (*cit)->key()) {
        propFound = true;
        PropertyListItem* pli = dynamic_cast<PropertyListItem*>(it->first);
        if (pli) {
          (*cit)->setValue(it->first->currentItem());
        }
        else {
          (*cit)->setValue(it->first->value());
        }
        break;
      }
    }
    if (!propFound) {
      throw BInternalException(__FILE__, __LINE__,
        tr("Property '%1' reguested but not found").arg((*cit)->key()));
    }
  }
}

bool PropertyList::addPropertyItem( PropertyItem *&item, BProperty* property )
{
  switch ( property->type() ) {
  case BProperty::IntType:
    item = new PropertyIntItem( this, item, 0,property->name(), TRUE );
    break;
  case BProperty::UIntType:
    item = new PropertyIntItem( this, item, 0,property->name(), FALSE );
    break;
  case BProperty::DoubleType: {
    item = new PropertyDoubleItem( this, item, 0,property->name());
    QVariant op = property->options();
    if (op.canCast(QVariant::List)) {
      QValueList<QVariant> op1 = op.asList();
      double bottom = 0;
      double top = 0;
      int decimals = 4;
      bool setRange = true;
      QValueList<QVariant>::iterator it = op1.at(0);
      if (it != op1.end()) {
        bottom = (*it).asDouble();
      }
      else {
        setRange = false;
      }
      it = op1.at(1);
      if (it != op1.end()) {
        top = (*it).asDouble();
      }
      else {
        setRange = false;
      }
      it = op1.at(3);
      if (it != op1.end()) {
        decimals = (*it).asInt();
      }
      if (setRange) {
        dynamic_cast<PropertyDoubleItem*>(item)->setRange(bottom, top, decimals);
      }
    }
    }
    break;
  case BProperty::StringType:
    item = new PropertyTextItem( this, item, 0,property->name(), 
      FALSE, FALSE, FALSE);
    break;
  case BProperty::CourseType:
    item = new BPropertyCourseItem( this, item, 0,property->name());
    break;
  case BProperty::LongitudeType:
    item = new BPropertyCoordinateItem( this, item, 0,property->name(),
      BSeaMap::longitude);
    break;
  case BProperty::LatitudeType:
    item = new BPropertyCoordinateItem( this, item, 0,property->name(),
      BSeaMap::latitude);
    break;
  case BProperty::StringListType:
    item = new PropertyListItem( this, item, 0,property->name(), false);
    break;
  default:
    return FALSE;
  }
  if (item) {
    item->setEditable(property->isEditable());
    item->setWhatsThisText(property->whatsThisText());
  }
  return TRUE;
}

void PropertyList::paintEmptyArea( QPainter *p, const QRect &r )
{
  p->fillRect( r, *backColor2 );
}

void PropertyList::setCurrentItem( QListViewItem *i )
{
  if ( !i )
    return;
  
  if ( currentItem() )
    ( (PropertyItem*)currentItem() )->hideEditor();
  QListView::setCurrentItem( i );
  ( (PropertyItem*)currentItem() )->showEditor();
}

void PropertyList::updateEditorSize()
{
  if ( currentItem() )
    ( ( PropertyItem* )currentItem() )->showEditor();
}

/*!  This has to be called if the value if \a i should be set as
property to the currently edited widget.
*/

void PropertyList::valueChanged( PropertyItem *i )
{
  PropertyListItem* pli = dynamic_cast<PropertyListItem*>(i);
  QVariant vv = i->value();
  if (pli) {
    vv = i->currentItem();
  }
  BProperty prop(_items[i]->type(), _items[i]->key(), 
    _items[i]->name(), vv);
  emit valueChanged(prop);
  //    if ( !editor->widget() )
  //	return;
  /*    QString pn( tr( "Set '%1' of '%2'" ).arg( i->name() ).arg( editor->widget()->name() ) );
  SetPropertyCommand *cmd = new SetPropertyCommand( pn, editor->formWindow(),
  editor->widget(), editor,
  i->name(), WidgetFactory::property( editor->widget(), i->name() ),
  i->value(), i->currentItem(), i->currentItemFromObject() );
  
    cmd->execute();
    editor->formWindow()->commandHistory()->addCommand( cmd, TRUE );
  */
}

void PropertyList::layoutInitValue( PropertyItem * /* i */, bool /* changed */ )
{
/*    if ( !editor->widget() )
return;
QString pn( tr( "Set '%1' of '%2'" ).arg( i->name() ).arg( editor->widget()->name() ) );
SetPropertyCommand *cmd = new SetPropertyCommand( pn, editor->formWindow(),
editor->widget(), editor,
i->name(), WidgetFactory::property( editor->widget(), i->name() ),
i->value(), i->currentItem(), i->currentItemFromObject() );
cmd->execute();
if ( i->value().toString() != "-1" )
changed = TRUE;
i->setChanged( changed );
  */
}

void PropertyList::itemPressed( QListViewItem *i, const QPoint &p, int c )
{
  if ( !i )
    return;
  PropertyItem *pi = (PropertyItem*)i;
  if ( !pi->hasSubItems() )
    return;
  
  if ( c == 0 && viewport()->mapFromGlobal( p ).x() < 20 )
    toggleOpen( i );
}

void PropertyList::toggleOpen( QListViewItem *i )
{
  if ( !i )
    return;
  PropertyItem *pi = (PropertyItem*)i;
  if ( pi->hasSubItems() ) {
    pi->setOpen( !pi->isOpen() );
  } else {
    pi->toggle();
  }
}

bool PropertyList::eventFilter( QObject *o, QEvent *e )
{
  if ( !o || !e )
    return TRUE;
  
  PropertyItem *i = (PropertyItem*)currentItem();
  if ( e->type() == QEvent::KeyPress )
    theLastEvent = KeyEvent;
  else if ( e->type() == QEvent::MouseButtonPress )
    theLastEvent = MouseEvent;
  
  if ( o != this &&e->type() == QEvent::KeyPress ) {
    QKeyEvent *ke = (QKeyEvent*)e;
    if ( ( ke->key() == Key_Up || ke->key() == Key_Down ) &&
	     ( o != this || o != viewport() ) &&
       !( ke->state() & ControlButton ) ) {
      QApplication::sendEvent( this, (QKeyEvent*)e );
      return TRUE;
    } else if ( ( !::qt_cast<QLineEdit*>(o) ||
      ( ::qt_cast<QLineEdit*>(o) && ( (QLineEdit*)o )->isReadOnly() ) ) &&
		    i && i->hasSubItems() ) {
      if ( !i->isOpen() &&
        ( ke->key() == Key_Plus ||
        ke->key() == Key_Right ))
        i->setOpen( TRUE );
      else if ( i->isOpen() &&
		      ( ke->key() == Key_Minus ||
          ke->key() == Key_Left ) )
          i->setOpen( FALSE );
    } else if ( ( ke->key() == Key_Return || ke->key() == Key_Enter ) && ::qt_cast<QComboBox*>(o) ) {
      QKeyEvent ke2( QEvent::KeyPress, Key_Space, 0, 0 );
      QApplication::sendEvent( o, &ke2 );
      return TRUE;
    }
  } 
  
  /*    else if ( e->type() == QEvent::FocusOut && ::qt_cast<QLineEdit*>(o) && editor->formWindow() ) {
  QTimer::singleShot( 100, editor->formWindow()->commandHistory(), SLOT( checkCompressedCommand() ) );
  } 
  */    
  else if ( o == viewport() ) {
    QMouseEvent *me;
    PropertyListItem* i;
    switch ( e->type() ) {
    case QEvent::MouseButtonPress:
      me = (QMouseEvent*)e;
      i = (PropertyListItem*) itemAt( me->pos() );
      if ( i && ( ::qt_cast<PropertyColorItem*>(i) || ::qt_cast<PropertyPixmapItem*>(i) ) ) {
        pressItem = i;
        pressPos = me->pos();
        mousePressed = TRUE;
      }
      break;
    case QEvent::MouseMove:
      me = (QMouseEvent*)e;
      if ( me && me->state() & LeftButton && mousePressed) {
        
        i = (PropertyListItem*) itemAt( me->pos() );
        if ( i  && i == pressItem ) {
          
          if (( pressPos - me->pos() ).manhattanLength() > QApplication::startDragDistance() ){
            if ( ::qt_cast<PropertyColorItem*>(i) ) {
              QColor col = i->value().asColor();
              QColorDrag *drg = new QColorDrag( col, this );
              QPixmap pix( 25, 25 );
              pix.fill( col );
              QPainter p( &pix );
              p.drawRect( 0, 0, pix.width(), pix.height() );
              p.end();
              drg->setPixmap( pix );
              mousePressed = FALSE;
              drg->dragCopy();
            }
            else if ( ::qt_cast<PropertyPixmapItem*>(i) ) {
              QPixmap pix = i->value().asPixmap();
              if( !pix.isNull() ) {
                QImage img = pix.convertToImage();
                QImageDrag *drg = new QImageDrag( img, this );
                drg->setPixmap( pix );
                mousePressed = FALSE;
                drg->dragCopy();
              }
            }
          }
        }
      }
      break;
    default:
      break;
    }
  } else if ( o == header() ) {
    if ( e->type() == QEvent::ContextMenu ) {
      ((QContextMenuEvent *)e)->accept();
      QPopupMenu menu( 0 );
      menu.setCheckable( TRUE );
      const int cat_id = 1;
      const int alpha_id = 2;
      menu.insertItem( tr( "Sort &Categorized" ), cat_id );
      int alpha = menu.insertItem( tr( "Sort &Alphabetically" ), alpha_id );
      if ( showSorted )
        menu.setItemChecked( alpha_id, TRUE );
      else
        menu.setItemChecked( cat_id, TRUE );
      int res = menu.exec( ( (QContextMenuEvent*)e )->globalPos() );
      if ( res != -1 ) {
        bool newShowSorted = ( res == alpha );
        if ( showSorted != newShowSorted ) {
          showSorted = newShowSorted;
          //		    editor->clear();
          //		    editor->setup();
        }
      }
      return TRUE;
    }
  }
  
  return QListView::eventFilter( o, e );
}

/*!  This method re-initializes each item of the property list.
*/

void PropertyList::refetchData()
{
  QListViewItemIterator it( this );
  for ( ; it.current(); ++it ) {
    PropertyItem *i = (PropertyItem*)it.current();
    if ( !i->propertyParent() )
      setPropertyValue( i );
    if ( i->hasSubItems() )
      i->initChildren();
    bool changed = true;//MetaDataBase::isPropertyChanged( editor->widget(), i->name() );
    if ( ( i->name() == "layoutSpacing" || i->name() == "layoutMargin" ) ) {
      if ( i->value().toString() != "-1" )
        i->setChanged( TRUE, FALSE );
      else
        i->setChanged( FALSE, FALSE );
    }
    else if ( changed != i->isChanged() )
      i->setChanged( changed, FALSE );
  }
  updateEditorSize();
}

static void clearAlignList( QStrList &l )
{
  if ( l.count() == 1 )
    return;
  if ( l.find( "AlignAuto" ) != -1 )
    l.remove( "AlignAuto" );
  if ( l.find( "WordBreak" ) != -1 )
    l.remove( "WordBreak" );
}

/*!  This method initializes the value of the item \a i to the value
of the corresponding property.
*/

void PropertyList::setPropertyValue( PropertyItem * /* i */ )
{
}

void PropertyList::setCurrentProperty( const QString &n )
{
  if ( currentItem() && currentItem()->text( 0 ) == n ||
    currentItem() && ( (PropertyItem*)currentItem() )->propertyParent() &&
    ( (PropertyItem*)currentItem() )->propertyParent()->text( 0 ) == n )
    return;
  
  QListViewItemIterator it( this );
  for ( ; it.current(); ++it ) {
    if ( it.current()->text( 0 ) == n ) {
      setCurrentItem( it.current() );
      break;
    }
  }
}

void PropertyList::clear() {
  _items.clear();
  _properties.clear();
  QListView::clear();
}
/*PropertyEditor *PropertyList::propertyEditor() const
{
return editor;
}
*/

void PropertyList::resetProperty()
{
  if ( !currentItem() )
    return;
  PropertyItem *i = (PropertyItem*)currentItem();
  BProperty* p = _items.find(i)->second;
  i->setValue(p->value());
  i->setChanged(false, false);
  if ( i->hasSubItems() )
    i->initChildren();
  emit valueChanged(*p);
}

void PropertyList::viewportDragEnterEvent( QDragEnterEvent *e )
{
  PropertyListItem *i = (PropertyListItem*) itemAt( e->pos() );
  if( !i ) {
    e->ignore();
    return;
  }
  
  if ( ::qt_cast<PropertyColorItem*>(i) && QColorDrag::canDecode( e ) )
    e->accept();
  else if ( ::qt_cast<PropertyPixmapItem*>(i) && QImageDrag::canDecode( e ) )
    e->accept();
  else
    e->ignore();
}

void PropertyList::viewportDragMoveEvent ( QDragMoveEvent *e )
{
  PropertyListItem *i = (PropertyListItem*) itemAt( e->pos() );
  if( !i ) {
    e->ignore();
    return;
  }
  
  if ( ::qt_cast<PropertyColorItem*>(i) && QColorDrag::canDecode( e ) )
    e->accept();
  else if ( ::qt_cast<PropertyPixmapItem*>(i) && QImageDrag::canDecode( e ) )
    e->accept();
  else
    e->ignore();
}

void PropertyList::viewportDropEvent ( QDropEvent *e )
{
  PropertyListItem *i = (PropertyListItem*) itemAt( e->pos() );
  if( !i ) {
    e->ignore();
    return;
  }
  
  if ( ::qt_cast<PropertyColorItem*>(i) && QColorDrag::canDecode( e ) ) {
    QColor color;
    QColorDrag::decode( e, color );
    i->setValue( QVariant( color ) );
    valueChanged( i );
    e->accept();
  }
  else if ( ::qt_cast<PropertyPixmapItem*>(i)  && QImageDrag::canDecode( e ) ) {
    QImage img;
    QImageDrag::decode( e, img );
    QPixmap pm;
    pm.convertFromImage( img );
    i->setValue( QVariant( pm ) );
    valueChanged( i );
    e->accept();
  }
  else
    e->ignore();
}

QString PropertyList::whatsThisAt( const QPoint &p )
{
  return whatsThisText( itemAt( p ) );
}

void PropertyList::showCurrentWhatsThis()
{
  if ( !currentItem() )
    return;
  QPoint p( 0, currentItem()->itemPos() );
  p = viewport()->mapToGlobal( contentsToViewport( p ) );
  QWhatsThis::display( whatsThisText( currentItem() ), p, viewport() );
}

QString PropertyList::whatsThisText( QListViewItem *i )
{
  if ( !i )
    return QString::null;
  return ((PropertyItem*)i)->whatsThisText();
}

void PropertyList::readPropertyDocs()
{
/*    if ( !propertyDocs.isEmpty() )
return;

  QString docFile = MainWindow::self->documentationPath() + "/propertydocs";
  QFile f( docFile );
  if ( !f.open( IO_ReadOnly ) )
  return;
  QDomDocument doc;
  QString errMsg;
  int errLine;
  if ( !doc.setContent( &f, &errMsg, &errLine ) )
  return;
  QDomElement e = doc.firstChild().toElement().firstChild().toElement();
  
    for ( ; !e.isNull(); e = e.nextSibling().toElement() ) {
    QDomElement n = e.firstChild().toElement();
    QString name;
    QString doc;
    for ( ; !n.isNull(); n = n.nextSibling().toElement() ) {
    if ( n.tagName() == "name" )
    name = n.firstChild().toText().data();
    else if ( n.tagName() == "doc" )
    doc = n.firstChild().toText().data();
    }
    doc.insert( 0, "<p><b>" + name + "</b></p>" );
    propertyDocs.insert( name, doc );
    }
  */
}

PropertyList::LastEventType PropertyList::lastEvent()
{
  return theLastEvent;
}


